import uuid
from typing import Any, Literal, Union, overload
from typing_extensions import TypeGuard


@overload
def is_valid_uuid(val: None) -> Literal[False]: ...


@overload
def is_valid_uuid(val: str) -> TypeGuard[str]: ...


@overload
def is_valid_uuid(val: uuid.UUID) -> TypeGuard[uuid.UUID]: ...


@overload
def is_valid_uuid(val: Any) -> Literal[False]: ...


def is_valid_uuid(val: Any) -> TypeGuard[Union[str, uuid.UUID]]:
    try:
        uuid.UUID(str(val))
        return True
    except (ValueError, TypeError):
        return False



from typing import Any, Callable, Dict, List, Optional, Sequence, Union

from pydantic import ConfigDict, Field, SerializationInfo, field_serializer, model_serializer

from vellum.client.core.pydantic_utilities import UniversalBaseModel
from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.mocks import MockNodeExecution
from vellum.workflows.outputs.base import BaseOutputs
from vellum.workflows.references.constant import ConstantValueReference
from vellum.workflows.triggers import BaseTrigger


class DatasetRow(UniversalBaseModel):
    """
    Universal base model representing a dataset row with a label and inputs.

    Attributes:
        id: Optional unique identifier for the dataset row
        label: String label for the dataset row
        inputs: BaseInputs instance or dict containing the input data
        workflow_trigger: Optional Trigger instance for this scenario
        mocks: Optional sequence of node output mocks for testing scenarios
    """

    model_config = ConfigDict(populate_by_name=True, arbitrary_types_allowed=True)

    id: Optional[str] = None
    label: str
    inputs: Union[BaseInputs, Dict[str, Any]] = Field(default_factory=BaseInputs)
    workflow_trigger: Optional[BaseTrigger] = None
    mocks: Optional[Sequence[Union[BaseOutputs, MockNodeExecution]]] = None

    @model_serializer(mode="wrap")
    def serialize_full_model(self, handler: Callable[[Any], Any], info: SerializationInfo) -> Dict[str, Any]:
        """Serialize the model and add node_id field computed from then_outputs."""
        serialized = handler(self)
        if not isinstance(serialized, dict):
            return serialized

        if "mocks" in serialized and serialized.get("mocks") is None:
            serialized.pop("mocks")

        if "workflow_trigger" in serialized:
            value = serialized.pop("workflow_trigger")
            if value is not None:
                serialized["workflow_trigger_id"] = value

        if "id" in serialized and serialized.get("id") is None:
            serialized.pop("id")

        return serialized

    @field_serializer("workflow_trigger")
    def serialize_workflow_trigger(self, workflow_trigger: Optional[BaseTrigger]) -> Optional[str]:
        """
        Custom serializer for workflow_trigger that converts it to a string ID.

        Args:
            workflow_trigger: Optional workflow trigger instance

        Returns:
            String representation of the trigger ID, or None if no trigger
        """
        if workflow_trigger is None:
            return None

        return str(workflow_trigger.__class__.__id__)

    @field_serializer("inputs")
    def serialize_inputs(self, inputs: Union[BaseInputs, Dict[str, Any]]) -> Dict[str, Any]:
        """
        Custom serializer for inputs that converts it to a dictionary.

        Args:
            inputs: Either a BaseInputs instance or dict to serialize

        Returns:
            Dictionary representation of the inputs
        """
        if isinstance(inputs, dict):
            return inputs

        result = {}
        for input_descriptor, value in inputs:
            if not input_descriptor.name.startswith("__"):
                result[input_descriptor.name] = value

        return result

    @field_serializer("mocks")
    def serialize_mocks(
        self, mocks: Optional[Sequence[Union[BaseOutputs, MockNodeExecution]]], info
    ) -> Optional[List[Dict[str, Any]]]:
        """
        Custom serializer for mocks that normalizes both BaseOutputs and MockNodeExecution
        to a consistent dict format with node_id, when_condition, and then_outputs.

        Args:
            mocks: Optional sequence of BaseOutputs or MockNodeExecution instances
            info: Serialization info containing context

        Returns:
            List of normalized mock execution dicts, or None if input is None
        """
        if mocks is None:
            return None

        result = []
        for mock in mocks:
            if isinstance(mock, MockNodeExecution):
                mock_exec = mock
            else:
                mock_exec = MockNodeExecution(
                    when_condition=ConstantValueReference(True),
                    then_outputs=mock,
                )

            result.append(mock_exec.model_dump(context=info.context))

        return result



import pytest
from typing import Any, Union

from vellum.workflows.utils.validate import validate_target_type


@pytest.mark.parametrize(
    ["declared_type", "target_type"],
    [
        (str, int),
        (list[str], list[int]),
        (str, Union[str, int]),
        (int, Union[str, int]),
        (list[str], Union[list[str], list[int]]),
        (Union[str, int], Union[bool, float]),
    ],
    ids=[
        "str_int",
        "list_str_int",
        "str_union_str_int",
        "int_union_str_int",
        "list_str_union",
        "union_str_int_union_bool_float",
    ],
)
def test_validate__should_raise_exception(
    declared_type,
    target_type,
):
    """
    Tests that validate_target_type raises an exception for mismatched types.
    """

    # WHEN validating the target type
    with pytest.raises(ValueError) as exc_info:
        validate_target_type(declared_type, target_type)

    # THEN an exception should be raised
    assert "Output type mismatch" in str(exc_info.value)


@pytest.mark.parametrize(
    ["declared_type", "target_type"],
    [
        (str, str),
        (list[str], list[str]),
        (dict, dict[str, str]),
        (str, Any),
        (Any, str),
        (Union[str, int], str),
        (Union[str, int], int),
        (Union[str, int], Union[str, int]),
        (Union[str, int], Union[int, str]),
        (Union[str, int], Union[str, bool]),
    ],
    ids=[
        "str",
        "list_str",
        "bare_dict_params_dict",
        "str_any",
        "any_str",
        "union_str_int_str",
        "union_str_int_int",
        "union_str_int_union_str_int",
        "union_str_int_union_int_str",
        "union_str_int_union_str_bool",
    ],
)
def test_validate__should_validate(
    declared_type,
    target_type,
):
    """
    Tests that validate_target_type accepts matching types.
    """

    # WHEN validating the target type
    # THEN no exception should be raised
    validate_target_type(declared_type, target_type)



from deepdiff import DeepDiff

from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_error_node.workflow import BasicErrorNodeWorkflow


def test_serialize_workflow():
    # GIVEN a Workflow with an error node
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=BasicErrorNodeWorkflow)

    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 1
    assert not DeepDiff(
        [
            {
                "id": "5d9edd44-b35b-4bad-ad51-ccdfe8185ff5",
                "key": "threshold",
                "type": "NUMBER",
                "default": None,
                "required": True,
                "extensions": {"color": None},
            }
        ],
        input_variables,
        ignore_order=True,
    )

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 1
    assert not DeepDiff(
        [
            {
                "id": "04c5c6be-f5e1-41b8-b668-39e179790d9e",
                "key": "final_value",
                "type": "NUMBER",
            }
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]

    # AND the error node should be serialized correctly
    error_node = next(n for n in workflow_raw_data["nodes"] if (n.get("base") or {}).get("name") == "ErrorNode")

    assert error_node["id"] == "98597c0e-8951-4320-98ed-efd7f79c00cf"
    assert error_node["type"] == "ERROR"

    assert error_node["base"]["name"] == "ErrorNode"
    assert error_node["base"]["module"] == ["vellum", "workflows", "nodes", "core", "error_node", "node"]

    assert error_node["definition"]["name"] == "FailNode"
    assert error_node["definition"]["module"] == ["tests", "workflows", "basic_error_node", "workflow"]

    assert error_node["trigger"]["id"] == "865dcf9c-8fe8-44c0-bf36-2c92eb47927e"
    assert error_node["trigger"]["merge_behavior"] == "AWAIT_ATTRIBUTES"

    assert error_node["ports"] == []

    assert error_node["data"]["label"] == "Fail Node"
    assert error_node["data"]["target_handle_id"] == "865dcf9c-8fe8-44c0-bf36-2c92eb47927e"
    assert error_node["data"]["error_source_input_id"] == "f7fc097f-db5b-48c1-8c33-391834678521"

    assert len(error_node["inputs"]) == 1
    assert error_node["inputs"][0]["key"] == "error_source_input_id"

    passthrough_nodes = [node for node in workflow_raw_data["nodes"] if node["type"] == "GENERIC"]
    assert len(passthrough_nodes) == 2



import pytest
from datetime import datetime
from unittest.mock import MagicMock
from typing import List

from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.inputs.dataset_row import DatasetRow
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.sandbox import WorkflowSandboxRunner
from vellum.workflows.state.base import BaseState
from vellum.workflows.triggers import ScheduleTrigger
from vellum.workflows.workflows.base import BaseWorkflow


@pytest.fixture
def mock_logger(mocker):
    return mocker.patch("vellum.workflows.sandbox.load_logger")


@pytest.mark.parametrize(
    ["run_kwargs", "expected_last_log"],
    [
        ({}, "final_results: first"),
        ({"index": 1}, "final_results: second"),
        ({"index": -4}, "final_results: first"),
        ({"index": 100}, "final_results: second"),
    ],
    ids=["default", "specific", "negative", "out_of_bounds"],
)
def test_sandbox_runner__happy_path(mock_logger, run_kwargs, expected_last_log):
    # GIVEN we capture the logs to stdout
    logs = []
    mock_logger.return_value.info.side_effect = lambda msg: logs.append(msg)

    # AND an example workflow
    class Inputs(BaseInputs):
        foo: str

    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            bar = Inputs.foo

    class Workflow(BaseWorkflow[Inputs, BaseState]):
        graph = StartNode

        class Outputs(BaseWorkflow.Outputs):
            final_results = StartNode.Outputs.bar

    # AND a dataset for this workflow
    inputs: List[Inputs] = [
        Inputs(foo="first"),
        Inputs(foo="second"),
    ]

    # WHEN we run the sandbox
    runner = WorkflowSandboxRunner(workflow=Workflow(), inputs=inputs)
    runner.run(**run_kwargs)

    # THEN we see the logs
    assert logs == [
        "Just started Node: StartNode",
        "Just finished Node: StartNode",
        "Workflow fulfilled!",
        "----------------------------------",
        expected_last_log,
    ]


def test_sandbox_runner_with_dict_inputs(mock_logger):
    """
    Test that WorkflowSandboxRunner can run with dict inputs in DatasetRow.
    """

    # GIVEN we capture the logs to stdout
    logs = []
    mock_logger.return_value.info.side_effect = lambda msg: logs.append(msg)

    class Inputs(BaseInputs):
        message: str

    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result = Inputs.message

    class Workflow(BaseWorkflow[Inputs, BaseState]):
        graph = StartNode

        class Outputs(BaseWorkflow.Outputs):
            final_output = StartNode.Outputs.result

    dataset = [
        DatasetRow(label="test_row", inputs={"message": "Hello from dict"}),
    ]

    # WHEN we run the sandbox with the DatasetRow containing dict inputs
    runner = WorkflowSandboxRunner(workflow=Workflow(), dataset=dataset)
    runner.run()

    assert logs == [
        "Just started Node: StartNode",
        "Just finished Node: StartNode",
        "Workflow fulfilled!",
        "----------------------------------",
        "final_output: Hello from dict",
    ]


def test_sandbox_runner_with_workflow_trigger(mock_logger):
    """
    Test that WorkflowSandboxRunner can run with DatasetRow containing workflow_trigger.
    """

    # GIVEN we capture the logs to stdout
    logs = []
    mock_logger.return_value.info.side_effect = lambda msg: logs.append(msg)

    class MySchedule(ScheduleTrigger):
        class Config(ScheduleTrigger.Config):
            cron = "* * * * *"
            timezone = "UTC"

    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result = MySchedule.current_run_at

    class Workflow(BaseWorkflow):
        graph = MySchedule >> StartNode

        class Outputs(BaseWorkflow.Outputs):
            final_output = StartNode.Outputs.result

    # AND a trigger instance
    trigger_instance = MySchedule(current_run_at=datetime.min, next_run_at=datetime.now())

    # AND a dataset with workflow_trigger instance
    dataset = [
        DatasetRow(
            label="test_row",
            inputs={"current_run_at": datetime.min, "next_run_at": datetime.now()},
            workflow_trigger=trigger_instance,
        ),
    ]

    # WHEN we run the sandbox with the DatasetRow containing workflow_trigger
    runner = WorkflowSandboxRunner(workflow=Workflow(), dataset=dataset)
    runner.run()

    # THEN the workflow should run successfully
    assert logs == [
        "Just started Node: StartNode",
        "Just finished Node: StartNode",
        "Workflow fulfilled!",
        "----------------------------------",
        "final_output: 0001-01-01 00:00:00",
    ]

    # AND the dataset row should have the trigger instance
    assert dataset[0].workflow_trigger == trigger_instance
    assert isinstance(dataset[0].workflow_trigger, MySchedule)


def test_sandbox_runner_with_trigger_instance(mock_logger):
    """
    Test that WorkflowSandboxRunner can run with DatasetRow containing trigger instance.
    """

    # GIVEN we capture the logs to stdout
    logs = []
    mock_logger.return_value.info.side_effect = lambda msg: logs.append(msg)

    # AND a trigger class
    class MySchedule(ScheduleTrigger):
        class Config(ScheduleTrigger.Config):
            cron = "* * * * *"
            timezone = "UTC"

    # AND a workflow that uses the trigger
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result = MySchedule.current_run_at

    class Workflow(BaseWorkflow):
        graph = MySchedule >> StartNode

        class Outputs(BaseWorkflow.Outputs):
            final_output = StartNode.Outputs.result

    # AND a trigger instance
    trigger_instance = MySchedule(current_run_at=datetime.min, next_run_at=datetime.now())

    # AND a dataset with trigger instance
    dataset = [
        DatasetRow(
            label="test_row_with_instance",
            inputs={"current_run_at": datetime.min, "next_run_at": datetime.now()},
            workflow_trigger=trigger_instance,
        ),
    ]

    # WHEN we run the sandbox with the DatasetRow containing trigger instance
    runner = WorkflowSandboxRunner(workflow=Workflow(), dataset=dataset)
    runner.run()

    # THEN the workflow should run successfully
    assert logs == [
        "Just started Node: StartNode",
        "Just finished Node: StartNode",
        "Workflow fulfilled!",
        "----------------------------------",
        "final_output: 0001-01-01 00:00:00",
    ]

    # AND the dataset row should have the trigger instance
    assert dataset[0].workflow_trigger == trigger_instance
    assert isinstance(dataset[0].workflow_trigger, MySchedule)


def test_dataset_row_serialization_with_workflow_trigger():
    """
    Test that DatasetRow serializes workflow_trigger field to workflow_trigger_id.
    """

    # GIVEN a trigger class
    class MySchedule(ScheduleTrigger):
        class Config(ScheduleTrigger.Config):
            cron = "* * * * *"
            timezone = "UTC"

    # AND a trigger instance
    trigger_instance = MySchedule(current_run_at=datetime.min, next_run_at=datetime.now())

    # AND a DatasetRow constructed with workflow_trigger
    dataset_row = DatasetRow(
        label="test_serialization",
        inputs={"foo": "bar"},
        workflow_trigger=trigger_instance,
    )

    # WHEN we serialize the DatasetRow
    serialized = dataset_row.model_dump()

    # THEN the serialized dict should contain workflow_trigger_id
    assert "workflow_trigger_id" in serialized
    assert serialized["workflow_trigger_id"] == str(MySchedule.__id__)


def test_sandbox_runner_with_node_output_mocks(mock_logger, mocker):
    """
    Tests that WorkflowSandboxRunner passes mocks from DatasetRow to workflow.stream().
    """

    class Inputs(BaseInputs):
        message: str

    class TestNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result: str

    class Workflow(BaseWorkflow[Inputs, BaseState]):
        graph = TestNode

        class Outputs(BaseWorkflow.Outputs):
            final_output = TestNode.Outputs.result

    mock_outputs = TestNode.Outputs(result="mocked_result")

    # AND a dataset with mocks
    dataset = [
        DatasetRow(
            label="test_with_mocks",
            inputs={"message": "test"},
            mocks=[mock_outputs],
        ),
    ]

    workflow_instance = Workflow()
    original_stream = workflow_instance.stream
    stream_mock = MagicMock(return_value=original_stream(inputs=Inputs(message="test")))
    mocker.patch.object(workflow_instance, "stream", stream_mock)

    # WHEN we run the sandbox with the DatasetRow containing mocks
    runner = WorkflowSandboxRunner(workflow=workflow_instance, dataset=dataset)
    runner.run()

    stream_mock.assert_called_once()
    call_kwargs = stream_mock.call_args.kwargs
    assert "node_output_mocks" in call_kwargs
    assert call_kwargs["node_output_mocks"] == [mock_outputs]



from deepdiff import DeepDiff

from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_generic_node.workflow import BasicGenericNodeWorkflow


def test_serialize_workflow():
    # GIVEN a Workflow that uses a generic node
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=BasicGenericNodeWorkflow)

    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 1
    assert not DeepDiff(
        [
            {
                "id": "a07c2273-34a7-42b5-bcad-143b6127cc8a",
                "key": "input",
                "type": "STRING",
                "default": None,
                "required": True,
                "extensions": {"color": None},
            },
        ],
        input_variables,
        ignore_order=True,
    )

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 1
    assert not DeepDiff(
        [
            {"id": "2b6389d0-266a-4be4-843e-4e543dd3d727", "key": "output", "type": "STRING"},
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]

    # AND the generic node should be serialized correctly
    generic_node = next(n for n in workflow_raw_data["nodes"] if (n.get("base") or {}).get("name") == "BaseNode")
    assert generic_node["id"] == "bf98371c-65d3-43c1-99a2-0f5369397847"

    # AND the nodes should have been auto-positioned since they all started at (0,0)
    nodes = workflow_raw_data["nodes"]
    positions = [node["display_data"]["position"] for node in nodes]

    nodes_at_zero = sum(1 for pos in positions if pos["x"] == 0.0 and pos["y"] == 0.0)
    assert nodes_at_zero < len(nodes), "Auto layout should have positioned nodes away from (0,0)"

    x_positions = [pos["x"] for pos in positions]
    assert all(x >= 0 for x in x_positions), "All x positions should be non-negative"
    assert len(set(x_positions)) > 1 or len(nodes) == 1, "Nodes should be spread across different x positions"

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "BasicGenericNodeWorkflow",
        "module": ["tests", "workflows", "basic_generic_node", "workflow"],
    }



from deepdiff import DeepDiff

from vellum.workflows.nodes.utils import ADORNMENT_MODULE_NAME
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_code_execution_node.try_workflow import TrySimpleCodeExecutionWorkflow
from tests.workflows.basic_code_execution_node.workflow import SimpleCodeExecutionWithFilepathWorkflow
from tests.workflows.basic_code_execution_node.workflow_with_code import SimpleCodeExecutionWithCodeWorkflow


def test_serialize_workflow_with_filepath():
    # GIVEN a Workflow with a code execution node
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=SimpleCodeExecutionWithFilepathWorkflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 0

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 2
    assert not DeepDiff(
        [
            {"id": "1cee930f-342f-421c-89fc-ff212b3764bb", "key": "log", "type": "STRING"},
            {"id": "f6a3e3e0-f83f-4491-8b7a-b20fddd7160c", "key": "result", "type": "NUMBER"},
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]

    # AND each node should be serialized correctly

    code_execution_node = next(
        n for n in workflow_raw_data["nodes"] if (n.get("base") or {}).get("name") == "CodeExecutionNode"
    )
    assert code_execution_node == {
        "id": "b4a0526b-0e31-4f63-892a-6d5197d09acf",
        "type": "CODE_EXECUTION",
        "inputs": [
            {
                "id": "e7a6c15f-083f-4775-8315-22b8fb62b9ba",
                "key": "code",
                "value": {
                    "rules": [
                        {
                            "type": "CONSTANT_VALUE",
                            "data": {"type": "STRING", "value": "# flake8: noqa\ndef main():\n    return 0\n"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "1248f512-cc1b-4daa-a444-3bf7a1f1b8f9",
                "key": "runtime",
                "value": {
                    "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "STRING", "value": "PYTHON_3_11_6"}}],
                    "combinator": "OR",
                },
            },
        ],
        "data": {
            "label": "Simple Code Execution Node",
            "error_output_id": None,
            "source_handle_id": "6aed9e19-9d26-457b-966d-0a9112f84070",
            "target_handle_id": "f1ea9f65-e225-49cc-a779-6bd0797ba22a",
            "code_input_id": "e7a6c15f-083f-4775-8315-22b8fb62b9ba",
            "runtime_input_id": "1248f512-cc1b-4daa-a444-3bf7a1f1b8f9",
            "output_type": "NUMBER",
            "packages": [{"name": "openai", "version": "1.0.0"}],
            "output_id": "730c8f17-891c-40b4-b43d-26672bd38eef",
            "log_output_id": "8eea2893-1e79-4d1c-ba51-2e045968abfb",
        },
        "display_data": {"position": {"x": 200.0, "y": -50.0}},
        "base": {
            "module": [
                "vellum",
                "workflows",
                "nodes",
                "displayable",
                "code_execution_node",
                "node",
            ],
            "name": "CodeExecutionNode",
        },
        "definition": {
            "module": ["tests", "workflows", "basic_code_execution_node", "workflow"],
            "name": "SimpleCodeExecutionNode",
        },
        "trigger": {
            "id": "f1ea9f65-e225-49cc-a779-6bd0797ba22a",
            "merge_behavior": "AWAIT_ANY",
        },
        "ports": [{"id": "6aed9e19-9d26-457b-966d-0a9112f84070", "name": "default", "type": "DEFAULT"}],
        "outputs": [
            {
                "id": "730c8f17-891c-40b4-b43d-26672bd38eef",
                "name": "result",
                "schema": {"type": "integer"},
                "type": "NUMBER",
                "value": None,
            },
            {
                "id": "8eea2893-1e79-4d1c-ba51-2e045968abfb",
                "name": "log",
                "schema": {"type": "string"},
                "type": "STRING",
                "value": None,
            },
        ],
    }

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "SimpleCodeExecutionWithFilepathWorkflow",
        "module": [
            "tests",
            "workflows",
            "basic_code_execution_node",
            "workflow",
        ],
    }


def test_serialize_workflow_with_code():
    # GIVEN a Workflow with a code execution node
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=SimpleCodeExecutionWithCodeWorkflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 0

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 2
    assert not DeepDiff(
        [
            {"id": "283d6849-f3ed-4beb-b261-cf70f90e8d10", "key": "result", "type": "NUMBER"},
            {"id": "4c136180-050b-4422-a7a4-2a1c6729042c", "key": "log", "type": "STRING"},
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]

    # AND each node should be serialized correctly

    code_execution_node = next(
        n for n in workflow_raw_data["nodes"] if (n.get("base") or {}).get("name") == "CodeExecutionNode"
    )
    assert code_execution_node == {
        "id": "ae9c5da6-242e-4e0d-abe6-344e2ada3ce3",
        "type": "CODE_EXECUTION",
        "inputs": [
            {
                "id": "331db512-0127-46ad-aeed-1abacb287a9e",
                "key": "code",
                "value": {
                    "rules": [
                        {
                            "type": "CONSTANT_VALUE",
                            "data": {"type": "STRING", "value": 'def main() -> str:\n    return "Hello, World!"\n'},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "dd7e4f60-858a-457d-9ad3-50a25871650f",
                "key": "runtime",
                "value": {
                    "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "STRING", "value": "PYTHON_3_11_6"}}],
                    "combinator": "OR",
                },
            },
        ],
        "data": {
            "label": "Simple Code Execution Node",
            "error_output_id": None,
            "source_handle_id": "d2560e23-cbb6-4b73-b082-294982da72aa",
            "target_handle_id": "e4a738e8-bbcf-47d2-b7c9-6a034dd412f0",
            "code_input_id": "331db512-0127-46ad-aeed-1abacb287a9e",
            "runtime_input_id": "dd7e4f60-858a-457d-9ad3-50a25871650f",
            "output_type": "NUMBER",
            "packages": [],
            "output_id": "9b2b939b-c840-4e01-99c1-43a3418d7fb7",
            "log_output_id": "e3536bf9-cf37-4d16-b835-770a9ee6cc53",
        },
        "display_data": {"position": {"x": 200.0, "y": -50.0}},
        "base": {
            "name": "CodeExecutionNode",
            "module": ["vellum", "workflows", "nodes", "displayable", "code_execution_node", "node"],
        },
        "definition": {
            "name": "SimpleCodeExecutionNode",
            "module": ["tests", "workflows", "basic_code_execution_node", "workflow_with_code"],
        },
        "trigger": {
            "id": "e4a738e8-bbcf-47d2-b7c9-6a034dd412f0",
            "merge_behavior": "AWAIT_ANY",
        },
        "ports": [{"id": "d2560e23-cbb6-4b73-b082-294982da72aa", "name": "default", "type": "DEFAULT"}],
        "outputs": [
            {
                "id": "9b2b939b-c840-4e01-99c1-43a3418d7fb7",
                "name": "result",
                "schema": {"type": "integer"},
                "type": "NUMBER",
                "value": None,
            },
            {
                "id": "e3536bf9-cf37-4d16-b835-770a9ee6cc53",
                "name": "log",
                "schema": {"type": "string"},
                "type": "STRING",
                "value": None,
            },
        ],
    }

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "SimpleCodeExecutionWithCodeWorkflow",
        "module": ["tests", "workflows", "basic_code_execution_node", "workflow_with_code"],
    }


def test_serialize_workflow__try_wrapped():
    # GIVEN a Workflow with a code execution node
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=TrySimpleCodeExecutionWorkflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 0

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 2
    assert not DeepDiff(
        [
            {"id": "5fbd27a0-9831-49c7-93c8-9c2a28c78696", "key": "log", "type": "STRING"},
            {"id": "400f9ffe-e700-4204-a810-e06123565947", "key": "result", "type": "NUMBER"},
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]

    # AND each node should be serialized correctly

    code_execution_node = next(
        n for n in workflow_raw_data["nodes"] if (n.get("base") or {}).get("name") == "CodeExecutionNode"
    )
    assert code_execution_node == {
        "id": "1c910367-dff1-4466-85bc-6a8ec4ca039d",
        "type": "CODE_EXECUTION",
        "inputs": [
            {
                "id": "5049de20-77c9-4bcf-a572-260ec18c0528",
                "key": "code",
                "value": {
                    "rules": [
                        {
                            "type": "CONSTANT_VALUE",
                            "data": {"type": "STRING", "value": "# flake8: noqa\ndef main():\n    return 0\n"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "167a0eb0-3bb7-4f9c-9517-15eda41b7b58",
                "key": "runtime",
                "value": {
                    "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "STRING", "value": "PYTHON_3_11_6"}}],
                    "combinator": "OR",
                },
            },
        ],
        "data": {
            "label": "Simple Code Execution Node",
            "error_output_id": "b466ebdc-db85-4807-bccd-94c4d1b97478",
            "source_handle_id": "dede45ee-e17a-447c-b1d3-ed0d29ff1057",
            "target_handle_id": "f4a58613-628e-4a6d-aeae-4f81cf96bbf4",
            "code_input_id": "5049de20-77c9-4bcf-a572-260ec18c0528",
            "runtime_input_id": "167a0eb0-3bb7-4f9c-9517-15eda41b7b58",
            "output_type": "NUMBER",
            "packages": [{"name": "openai", "version": "1.0.0", "repository": "test-repo"}],
            "output_id": "a483823a-e856-4ad3-ab3b-cac3c2961536",
            "log_output_id": "0022691d-4707-476f-a9f5-4947b5e36f07",
        },
        "display_data": {"position": {"x": 200.0, "y": -50.0}},
        "base": {
            "module": [
                "vellum",
                "workflows",
                "nodes",
                "displayable",
                "code_execution_node",
                "node",
            ],
            "name": "CodeExecutionNode",
        },
        "definition": {
            "module": [
                "tests",
                "workflows",
                "basic_code_execution_node",
                "try_workflow",
                "SimpleCodeExecutionNode",
                ADORNMENT_MODULE_NAME,
            ],
            "name": "TryNode",
        },
        "adornments": [
            {
                "id": "d602b4ed-fb9d-45ef-8d07-1b6c2d8cd893",
                "label": "Try Node",
                "base": {"name": "TryNode", "module": ["vellum", "workflows", "nodes", "core", "try_node", "node"]},
                "attributes": [
                    {
                        "id": "2a313b7c-accd-4af2-a05b-c73b73c222b7",
                        "name": "on_error_code",
                        "value": {"type": "CONSTANT_VALUE", "value": {"type": "JSON", "value": None}},
                    }
                ],
            }
        ],
        "trigger": {
            "id": "f4a58613-628e-4a6d-aeae-4f81cf96bbf4",
            "merge_behavior": "AWAIT_ANY",
        },
        "ports": [{"id": "dede45ee-e17a-447c-b1d3-ed0d29ff1057", "name": "default", "type": "DEFAULT"}],
        "outputs": [
            {
                "id": "a483823a-e856-4ad3-ab3b-cac3c2961536",
                "name": "result",
                "schema": {"type": "integer"},
                "type": "NUMBER",
                "value": None,
            },
            {
                "id": "0022691d-4707-476f-a9f5-4947b5e36f07",
                "name": "log",
                "schema": {"type": "string"},
                "type": "STRING",
                "value": None,
            },
        ],
    }

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "TrySimpleCodeExecutionWorkflow",
        "module": [
            "tests",
            "workflows",
            "basic_code_execution_node",
            "try_workflow",
        ],
    }



from copy import copy
from enum import Enum
import fnmatch
from functools import cached_property
import importlib
import inspect
import logging
import os
import pkgutil
import re
import traceback
from uuid import UUID
from typing import Any, Dict, ForwardRef, Generic, List, Optional, Set, Tuple, Type, TypeVar, Union, cast, get_args

from vellum.client import Vellum as VellumClient
from vellum.client.core.pydantic_utilities import UniversalBaseModel
from vellum.workflows import BaseWorkflow
from vellum.workflows.constants import undefined
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.edges import Edge
from vellum.workflows.edges.trigger_edge import TriggerEdge
from vellum.workflows.events.workflow import NodeEventDisplayContext, WorkflowEventDisplayContext
from vellum.workflows.exceptions import WorkflowInitializationException
from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.inputs.dataset_row import DatasetRow
from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.nodes.displayable.bases.utils import primitive_to_vellum_value
from vellum.workflows.nodes.displayable.final_output_node.node import FinalOutputNode
from vellum.workflows.nodes.utils import get_unadorned_node, get_unadorned_port, get_wrapped_node
from vellum.workflows.ports import Port
from vellum.workflows.references import OutputReference, StateValueReference, WorkflowInputReference
from vellum.workflows.triggers.integration import IntegrationTrigger
from vellum.workflows.triggers.manual import ManualTrigger
from vellum.workflows.triggers.schedule import ScheduleTrigger
from vellum.workflows.types.core import Json, JsonArray, JsonObject
from vellum.workflows.types.generics import WorkflowType
from vellum.workflows.types.utils import get_original_base
from vellum.workflows.utils.uuids import uuid4_from_hash
from vellum.workflows.utils.vellum_variables import primitive_type_to_vellum_variable_type
from vellum.workflows.vellum_client import create_vellum_client
from vellum_ee.workflows.display.base import (
    EdgeDisplay,
    EntrypointDisplay,
    StateValueDisplay,
    WorkflowInputsDisplay,
    WorkflowMetaDisplay,
    WorkflowOutputDisplay,
    WorkflowTriggerType,
    get_trigger_type_mapping,
)
from vellum_ee.workflows.display.editor.types import NodeDisplayData, NodeDisplayPosition
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.nodes.get_node_display_class import get_node_display_class
from vellum_ee.workflows.display.nodes.types import NodeOutputDisplay, PortDisplay
from vellum_ee.workflows.display.nodes.utils import raise_if_descriptor
from vellum_ee.workflows.display.types import (
    EdgeDisplays,
    EntrypointDisplays,
    NodeDisplays,
    NodeOutputDisplays,
    PortDisplays,
    StateValueDisplays,
    WorkflowDisplayContext,
    WorkflowInputsDisplays,
    WorkflowOutputDisplays,
)
from vellum_ee.workflows.display.utils.auto_layout import auto_layout_nodes
from vellum_ee.workflows.display.utils.exceptions import UserFacingException, WorkflowValidationError
from vellum_ee.workflows.display.utils.expressions import serialize_value
from vellum_ee.workflows.display.utils.metadata import (
    get_entrypoint_edge_id,
    get_regular_edge_id,
    get_trigger_edge_id,
    load_dataset_row_index_to_id_mapping,
)
from vellum_ee.workflows.display.utils.registry import register_workflow_display_class
from vellum_ee.workflows.display.utils.vellum import infer_vellum_variable_type
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

logger = logging.getLogger(__name__)

IGNORE_PATTERNS = [
    "*.pyc",
    "__pycache__",
    ".*",
    "node_modules/*",
    "*.log",
    "metadata.json",
]


class WorkflowSerializationError(UniversalBaseModel):
    message: str
    stacktrace: str


class WorkflowSerializationResult(UniversalBaseModel):
    exec_config: Dict[str, Any]
    errors: List[WorkflowSerializationError]
    dataset: Optional[List[Dict[str, Any]]] = None


BASE_MODULE_PATH = __name__


class _BaseWorkflowDisplayMeta(type):
    def __new__(mcs, name: str, bases: Tuple[Type[Any], ...], attrs: Dict[str, Any]) -> Type[Any]:
        cls = super().__new__(mcs, name, bases, attrs)

        # Automatically import all of the node displays now that we don't require the __init__.py file
        # to do so for us.
        module_path = cls.__module__
        if module_path.startswith(BASE_MODULE_PATH):
            return cls

        nodes_module_path = re.sub(r"\.workflow$", ".nodes", module_path)
        try:
            nodes_module = importlib.import_module(nodes_module_path)
        except Exception:
            # likely because there are no `.nodes` module in the display workflow's module path
            return cls

        if not hasattr(nodes_module, "__path__") or not hasattr(nodes_module, "__name__"):
            return cls

        for info in pkgutil.iter_modules(nodes_module.__path__, nodes_module.__name__ + "."):
            try:
                importlib.import_module(info.name)
            except Exception:
                continue

        return cls


class BaseWorkflowDisplay(Generic[WorkflowType], metaclass=_BaseWorkflowDisplayMeta):
    # Used to specify the display data for a workflow.
    workflow_display: Optional[WorkflowMetaDisplay] = None

    # Used to explicitly specify display data for a workflow's inputs.
    inputs_display: WorkflowInputsDisplays = {}

    # Used to explicitly specify display data for a workflow's state values.
    state_value_displays: StateValueDisplays = {}

    # Used to explicitly specify display data for a workflow's entrypoints.
    entrypoint_displays: EntrypointDisplays = {}

    # Used to explicitly specify display data for a workflow's outputs.
    output_displays: WorkflowOutputDisplays = {}

    # Used to explicitly specify display data for a workflow's edges.
    edge_displays: EdgeDisplays = {}

    # Used to explicitly specify display data for a workflow's ports.
    port_displays: PortDisplays = {}

    _serialized_files: List[str]

    _dry_run: bool

    def __init__(
        self,
        *,
        parent_display_context: Optional[WorkflowDisplayContext] = None,
        client: Optional[VellumClient] = None,
        dry_run: bool = False,
    ):
        self._parent_display_context = parent_display_context
        self._client = client or (
            # propagate the client from the parent display context if it is not provided
            self._parent_display_context.client
            if self._parent_display_context
            else create_vellum_client()
        )
        self._serialized_files = []
        self._dry_run = dry_run

    def serialize(self) -> JsonObject:
        try:
            self._workflow.validate()
        except WorkflowInitializationException as e:
            self.display_context.add_error(
                WorkflowValidationError(message=e.message, workflow_class_name=self._workflow.__name__)
            )

        self._serialized_files = [
            "__init__.py",
            "display/*",
            "inputs.py",
            "nodes/*",
            "state.py",
            "workflow.py",
            "triggers/*",
        ]

        input_variables: JsonArray = []
        for workflow_input_reference, workflow_input_display in self.display_context.workflow_input_displays.items():
            default = (
                primitive_to_vellum_value(workflow_input_reference.instance)
                if workflow_input_reference.instance is not None and workflow_input_reference.instance is not undefined
                else None
            )

            is_required = self._is_reference_required(workflow_input_reference)

            input_variables.append(
                {
                    "id": str(workflow_input_display.id),
                    "key": workflow_input_display.name or workflow_input_reference.name,
                    "type": infer_vellum_variable_type(workflow_input_reference),
                    "default": default.dict() if default else None,
                    "required": is_required,
                    "extensions": {"color": workflow_input_display.color},
                }
            )

        state_variables: JsonArray = []
        for state_value_reference, state_value_display in self.display_context.state_value_displays.items():
            default = (
                primitive_to_vellum_value(state_value_reference.instance)
                if state_value_reference.instance is not None and state_value_reference.instance is not undefined
                else None
            )

            is_required = self._is_reference_required(state_value_reference)

            state_variables.append(
                {
                    "id": str(state_value_display.id),
                    "key": state_value_display.name or state_value_reference.name,
                    "type": infer_vellum_variable_type(state_value_reference),
                    "default": default.dict() if default else None,
                    "required": is_required,
                    "extensions": {"color": state_value_display.color},
                }
            )

        serialized_nodes: Dict[UUID, JsonObject] = {}
        edges: JsonArray = []

        # Get all trigger edges from the workflow's subgraphs to check if trigger exists
        trigger_edges: List[TriggerEdge] = []
        for subgraph in self._workflow.get_subgraphs():
            trigger_edges.extend(list(subgraph.trigger_edges))

        # Determine if we need an ENTRYPOINT node and what ID to use
        manual_trigger_edges = [edge for edge in trigger_edges if issubclass(edge.trigger_class, ManualTrigger)]
        has_manual_trigger = len(manual_trigger_edges) > 0

        entrypoint_node_id: Optional[UUID] = None
        entrypoint_node_source_handle_id: Optional[UUID] = None

        if has_manual_trigger:
            # ManualTrigger: use trigger ID for ENTRYPOINT node (backward compatibility)
            trigger_class = manual_trigger_edges[0].trigger_class
            entrypoint_node_id = trigger_class.__id__
            entrypoint_node_source_handle_id = self.display_context.workflow_display.entrypoint_node_source_handle_id

            # Add ENTRYPOINT node for ManualTrigger workflows
            serialized_nodes[entrypoint_node_id] = {
                "id": str(entrypoint_node_id),
                "type": "ENTRYPOINT",
                "inputs": [],
                "data": {
                    "label": "Entrypoint Node",
                    "source_handle_id": str(entrypoint_node_source_handle_id),
                },
                "display_data": self.display_context.workflow_display.entrypoint_node_display.dict(),
                "base": None,
                "definition": None,
            }
        else:
            # All other cases: use workflow_display ENTRYPOINT node
            entrypoint_node_id = self.display_context.workflow_display.entrypoint_node_id
            entrypoint_node_source_handle_id = self.display_context.workflow_display.entrypoint_node_source_handle_id

            serialized_nodes[entrypoint_node_id] = {
                "id": str(entrypoint_node_id),
                "type": "ENTRYPOINT",
                "inputs": [],
                "data": {
                    "label": "Entrypoint Node",
                    "source_handle_id": str(entrypoint_node_source_handle_id),
                },
                "display_data": self.display_context.workflow_display.entrypoint_node_display.dict(),
                "base": None,
                "definition": None,
            }
        # else: has_only_integration_trigger without explicit entrypoint - no ENTRYPOINT node needed

        # Add all the nodes in the workflows
        for node in self._workflow.get_all_nodes():
            node_display = self.display_context.node_displays[node]

            try:
                try:
                    node.__validate__()
                except ValueError as validation_error:
                    # Only collect node validation errors directly to errors list, don't raise them
                    self.display_context.add_validation_error(validation_error)

                serialized_node = node_display.serialize(self.display_context)
            except (NotImplementedError, UserFacingException) as e:
                self.display_context.add_error(e)
                self.display_context.add_invalid_node(node)
                continue

            # Use wrapped node's ID as dict key for adornment wrappers to prevent overwrites
            wrapped_node = get_wrapped_node(node)
            if wrapped_node:
                wrapped_node_display = self.display_context.node_displays[wrapped_node]
                dict_key = wrapped_node_display.node_id
            else:
                dict_key = node_display.node_id

            serialized_nodes[dict_key] = serialized_node

        output_variables: JsonArray = []
        output_values: JsonArray = []
        final_output_nodes = [
            node for node in self.display_context.node_displays.keys() if issubclass(node, FinalOutputNode)
        ]
        final_output_node_outputs = {node.Outputs.value for node in final_output_nodes}
        unreferenced_final_output_node_outputs = final_output_node_outputs.copy()

        # Track the Workflow's output variables for each Workflow output
        for workflow_output, workflow_output_display in self.display_context.workflow_output_displays.items():
            inferred_type = infer_vellum_variable_type(workflow_output)
            # Remove the terminal node output from the unreferenced set
            if isinstance(workflow_output.instance, OutputReference):
                unreferenced_final_output_node_outputs.discard(workflow_output.instance)

            # Update the name of the terminal node if this output references a FinalOutputNode
            if workflow_output.instance in final_output_node_outputs:
                terminal_node_id = workflow_output.instance.outputs_class.__parent_class__.__id__
                serialized_terminal_node = serialized_nodes.get(terminal_node_id)
                if (
                    serialized_terminal_node
                    and "data" in serialized_terminal_node
                    and isinstance(serialized_terminal_node["data"], dict)
                ):
                    serialized_terminal_node["data"]["name"] = workflow_output_display.name

            try:
                output_value = serialize_value(self.workflow_id, self.display_context, workflow_output.instance)
            except UserFacingException as e:
                self.display_context.add_error(
                    UserFacingException(f"Failed to serialize output '{workflow_output.name}': {e}")
                )
                continue

            output_values.append(
                {
                    "output_variable_id": str(workflow_output_display.id),
                    "value": output_value,
                }
            )

            output_variables.append(
                {
                    "id": str(workflow_output_display.id),
                    "key": workflow_output_display.name,
                    "type": inferred_type,
                }
            )

        # If there are terminal nodes with no workflow output reference,
        # raise a serialization error
        if len(unreferenced_final_output_node_outputs) > 0:
            self.display_context.add_error(
                ValueError("Unable to serialize terminal nodes that are not referenced by workflow outputs.")
            )

        # Identify nodes that already have trigger edges so we can avoid duplicating entrypoint edges
        nodes_with_manual_trigger_edges: Set[Type[BaseNode]] = set()
        nodes_with_non_manual_trigger_edges: Set[Type[BaseNode]] = set()
        for trigger_edge in trigger_edges:
            try:
                unadorned_target_node = get_unadorned_node(trigger_edge.to_node)
            except Exception:
                continue

            if issubclass(trigger_edge.trigger_class, ManualTrigger):
                nodes_with_manual_trigger_edges.add(unadorned_target_node)
            else:
                nodes_with_non_manual_trigger_edges.add(unadorned_target_node)

        # Track nodes with explicit entrypoint overrides so we retain their edges even if they have triggers
        entrypoint_override_nodes: Set[Type[BaseNode]] = set()
        for entrypoint_node in self.entrypoint_displays.keys():
            try:
                entrypoint_override_nodes.add(get_unadorned_node(entrypoint_node))
            except Exception:
                continue

        # Determine which nodes have explicit non-trigger entrypoints in the graph
        non_trigger_entrypoint_nodes: Set[Type[BaseNode]] = set()
        for subgraph in self._workflow.get_subgraphs():
            # If the subgraph contains trigger edges, its entrypoints were derived from triggers
            if any(True for _ in subgraph.trigger_edges):
                continue
            for entrypoint in subgraph.entrypoints:
                try:
                    non_trigger_entrypoint_nodes.add(get_unadorned_node(entrypoint))
                except Exception:
                    continue

        # Add edges from entrypoint first to preserve expected ordering

        for target_node, entrypoint_display in self.display_context.entrypoint_displays.items():
            unadorned_target_node = get_unadorned_node(target_node)

            # Skip the auto-generated entrypoint edge when a manual trigger already targets this node or when a
            # non-manual trigger targets it without an explicit entrypoint override, unless the graph explicitly
            # defines a non-trigger entrypoint for it.
            has_manual_trigger = unadorned_target_node in nodes_with_manual_trigger_edges
            has_non_manual_trigger = unadorned_target_node in nodes_with_non_manual_trigger_edges
            has_override = unadorned_target_node in entrypoint_override_nodes
            if (
                has_manual_trigger or (has_non_manual_trigger and not has_override)
            ) and unadorned_target_node not in non_trigger_entrypoint_nodes:
                continue

            # Skip edges to invalid nodes
            if self._is_node_invalid(unadorned_target_node):
                continue

            if entrypoint_node_id is None:
                continue

            target_node_display = self.display_context.node_displays[unadorned_target_node]

            stable_edge_id = get_entrypoint_edge_id(unadorned_target_node, self._workflow.__module__)

            entrypoint_edge_dict: Dict[str, Json] = {
                "id": str(stable_edge_id) if stable_edge_id else str(entrypoint_display.edge_display.id),
                "source_node_id": str(entrypoint_node_id),
                "source_handle_id": str(entrypoint_node_source_handle_id),
                "target_node_id": str(target_node_display.node_id),
                "target_handle_id": str(target_node_display.get_trigger_id()),
                "type": "DEFAULT",
            }
            display_data = self._serialize_edge_display_data(entrypoint_display.edge_display)
            if display_data is not None:
                entrypoint_edge_dict["display_data"] = display_data
            edges.append(entrypoint_edge_dict)

        # Then add trigger edges
        for trigger_edge in trigger_edges:
            target_node = trigger_edge.to_node
            unadorned_target_node = get_unadorned_node(target_node)
            if issubclass(trigger_edge.trigger_class, ManualTrigger):
                nodes_with_manual_trigger_edges.add(unadorned_target_node)
            else:
                nodes_with_non_manual_trigger_edges.add(unadorned_target_node)

            # Skip edges to invalid nodes
            if self._is_node_invalid(unadorned_target_node):
                continue

            target_node_display = self.display_context.node_displays[unadorned_target_node]

            # Get the entrypoint display for this target node (if it exists)
            target_entrypoint_display = self.display_context.entrypoint_displays.get(target_node)
            if target_entrypoint_display is None:
                continue

            trigger_class = trigger_edge.trigger_class
            trigger_id = trigger_class.__id__

            # Determine source node ID and handle ID based on trigger type
            if issubclass(trigger_class, ManualTrigger):
                source_node_id = entrypoint_node_id
                source_handle_id = entrypoint_node_source_handle_id
            else:
                source_node_id = trigger_id
                source_handle_id = trigger_id

            # Prefer stable id from metadata mapping if present
            stable_edge_id = get_trigger_edge_id(trigger_class, unadorned_target_node, self._workflow.__module__)

            trigger_edge_dict: Dict[str, Json] = {
                "id": str(stable_edge_id) if stable_edge_id else str(target_entrypoint_display.edge_display.id),
                "source_node_id": str(source_node_id),
                "source_handle_id": str(source_handle_id),
                "target_node_id": str(target_node_display.node_id),
                "target_handle_id": str(target_node_display.get_trigger_id()),
                "type": "DEFAULT",
            }
            display_data = self._serialize_edge_display_data(target_entrypoint_display.edge_display)
            if display_data is not None:
                trigger_edge_dict["display_data"] = display_data
            edges.append(trigger_edge_dict)

        for (source_node_port, target_node), edge_display in self.display_context.edge_displays.items():
            unadorned_source_node_port = get_unadorned_port(source_node_port)
            unadorned_target_node = get_unadorned_node(target_node)

            # Skip edges that reference invalid nodes
            if self._is_node_invalid(unadorned_target_node) or self._is_node_invalid(
                unadorned_source_node_port.node_class
            ):
                continue

            source_node_port_display = self.display_context.port_displays[unadorned_source_node_port]
            target_node_display = self.display_context.node_displays[unadorned_target_node]

            stable_edge_id = get_regular_edge_id(
                unadorned_source_node_port.node_class,
                source_node_port_display.id,
                unadorned_target_node,
                self._workflow.__module__,
            )

            regular_edge_dict: Dict[str, Json] = {
                "id": str(stable_edge_id) if stable_edge_id else str(edge_display.id),
                "source_node_id": str(source_node_port_display.node_id),
                "source_handle_id": str(source_node_port_display.id),
                "target_node_id": str(target_node_display.node_id),
                "target_handle_id": str(
                    target_node_display.get_target_handle_id_by_source_node_id(source_node_port_display.node_id)
                ),
                "type": "DEFAULT",
            }
            display_data = self._serialize_edge_display_data(edge_display)
            if display_data is not None:
                regular_edge_dict["display_data"] = display_data
            edges.append(regular_edge_dict)

        nodes_list = list(serialized_nodes.values())
        nodes_dict_list = [cast(Dict[str, Any], node) for node in nodes_list if isinstance(node, dict)]

        all_nodes_at_zero = all(
            (
                isinstance(node.get("display_data"), dict)
                and isinstance(node["display_data"].get("position"), dict)
                and node["display_data"]["position"].get("x", 0) == 0.0
                and node["display_data"]["position"].get("y", 0) == 0.0
            )
            for node in nodes_dict_list
        )

        should_apply_auto_layout = all_nodes_at_zero and len(nodes_dict_list) > 0

        if should_apply_auto_layout:
            try:
                self._apply_auto_layout(nodes_dict_list, edges)
            except Exception as e:
                self.display_context.add_error(e)

        # Serialize workflow-level trigger if present
        triggers: Optional[JsonArray] = self._serialize_workflow_trigger()

        workflow_raw_data: JsonObject = {
            "nodes": cast(JsonArray, nodes_dict_list),
            "edges": edges,
            "display_data": self.display_context.workflow_display.display_data.dict(),
            "definition": {
                "name": self._workflow.__name__,
                "module": cast(JsonArray, self._workflow.__module__.split(".")),
            },
            "output_values": output_values,
        }

        result: JsonObject = {
            "workflow_raw_data": workflow_raw_data,
            "input_variables": input_variables,
            "state_variables": state_variables,
            "output_variables": output_variables,
        }

        if triggers is not None:
            result["triggers"] = triggers

        return result

    def _serialize_workflow_trigger(self) -> Optional[JsonArray]:
        """
        Serialize workflow-level trigger information.

        Returns:
            JsonArray with trigger data if a trigger is present, None otherwise.
            Each trigger in the array has: id (UUID), type (str), attributes (list)
        """
        # Get all trigger edges from the workflow's subgraphs
        trigger_edges = []
        for subgraph in self._workflow.get_subgraphs():
            trigger_edges.extend(list(subgraph.trigger_edges))

        if not trigger_edges:
            # No workflow-level trigger defined
            return None

        unique_trigger_classes = list(dict.fromkeys(edge.trigger_class for edge in trigger_edges))

        trigger_type_mapping = get_trigger_type_mapping()
        serialized_triggers: List[JsonObject] = []

        for trigger_class in unique_trigger_classes:
            # Get the trigger type from the mapping, or check if it's a subclass
            trigger_type = trigger_type_mapping.get(trigger_class)
            if trigger_type is None:
                # Check if it's a subclass of a known trigger type
                if issubclass(trigger_class, ManualTrigger):
                    trigger_type = WorkflowTriggerType.MANUAL
                elif issubclass(trigger_class, IntegrationTrigger):
                    trigger_type = WorkflowTriggerType.INTEGRATION
                elif issubclass(trigger_class, ScheduleTrigger):
                    trigger_type = WorkflowTriggerType.SCHEDULED
                else:
                    raise ValueError(
                        f"Unknown trigger type: {trigger_class.__name__}. "
                        f"Please add it to the trigger type mapping in get_trigger_type_mapping()."
                    )

            trigger_id = trigger_class.__id__

            # Serialize trigger attributes like node outputs
            trigger_data: JsonObject
            if trigger_type == WorkflowTriggerType.SCHEDULED:
                # For scheduled triggers, attributes should be empty
                # and cron/timezone should be top level

                config_class = trigger_class.Config
                cron_value = getattr(config_class, "cron", None)
                timezone_value = getattr(config_class, "timezone", None)

                trigger_data = {
                    "id": str(trigger_id),
                    "type": trigger_type.value,
                    "cron": cron_value,
                    "timezone": timezone_value,
                    "attributes": [],
                }
            else:
                # For other triggers, serialize attributes from attribute_references as VellumVariables
                attribute_references = trigger_class.attribute_references().values()
                trigger_attributes: JsonArray = cast(
                    JsonArray,
                    [
                        cast(
                            JsonObject,
                            {
                                "id": str(reference.id),
                                "key": reference.name,
                                "type": primitive_type_to_vellum_variable_type(reference),
                                "required": True,
                                "default": None,
                                "extensions": None,
                            },
                        )
                        for reference in sorted(attribute_references, key=lambda ref: ref.name)
                    ],
                )

                trigger_data = {
                    "id": str(trigger_id),
                    "type": trigger_type.value,
                    "attributes": trigger_attributes,
                }

                if trigger_type == WorkflowTriggerType.INTEGRATION and issubclass(trigger_class, IntegrationTrigger):
                    exec_config = self._serialize_integration_trigger_exec_config(trigger_class)
                    trigger_data["exec_config"] = exec_config

            # Serialize display_data from trigger's Display class
            display_class = trigger_class.Display
            display_data: JsonObject = {}

            # Add label if present
            if hasattr(display_class, "label") and display_class.label is not None:
                display_data["label"] = display_class.label

            # Add x and y coordinates if present
            if (
                hasattr(display_class, "x")
                and display_class.x is not None
                and hasattr(display_class, "y")
                and display_class.y is not None
            ):
                display_data["position"] = {
                    "x": display_class.x,
                    "y": display_class.y,
                }

            # Add z index if present
            if hasattr(display_class, "z_index") and display_class.z_index is not None:
                display_data["z_index"] = display_class.z_index

            # Add icon if present
            if hasattr(display_class, "icon") and display_class.icon is not None:
                display_data["icon"] = display_class.icon

            # Add color if present
            if hasattr(display_class, "color") and display_class.color is not None:
                display_data["color"] = display_class.color

            # Add comment if present
            if hasattr(display_class, "comment") and display_class.comment is not None:
                display_data["comment"] = {
                    "value": display_class.comment.value,
                    "expanded": display_class.comment.expanded,
                }

            # Don't include display_data for manual triggers
            if display_data and trigger_type != WorkflowTriggerType.MANUAL:
                trigger_data["display_data"] = display_data

            serialized_triggers.append(trigger_data)

        return cast(JsonArray, serialized_triggers)

    def _serialize_edge_display_data(self, edge_display: EdgeDisplay) -> Optional[JsonObject]:
        """Serialize edge display data, returning None if no display data is present."""
        if edge_display.z_index is not None:
            return {"z_index": edge_display.z_index}
        return None

    def _serialize_integration_trigger_exec_config(self, trigger_class: Type[IntegrationTrigger]) -> JsonObject:
        config_class = trigger_class.Config

        provider = getattr(config_class, "provider", None)
        if isinstance(provider, Enum):
            provider = provider.value
        elif provider is not None:
            provider = str(provider)
        slug = getattr(config_class, "slug", None)
        integration_name = getattr(config_class, "integration_name", None)

        setup_attributes: List[JsonObject] = []
        raw_setup_attributes = getattr(config_class, "setup_attributes", None)

        if isinstance(raw_setup_attributes, dict):
            for key, value in raw_setup_attributes.items():
                attribute_id = str(uuid4_from_hash(f"{trigger_class.__id__}|setup_attribute|{key}"))

                default_json: Optional[JsonObject] = None
                attribute_type = "STRING"

                if value is not None:
                    try:
                        vellum_value = primitive_to_vellum_value(value)
                        default_json = cast(JsonObject, self._model_dump(vellum_value))
                        attribute_type = cast(str, default_json.get("type", attribute_type))
                    except ValueError:
                        default_json = None

                setup_attributes.append(
                    cast(
                        JsonObject,
                        {
                            "id": attribute_id,
                            "key": str(key),
                            "type": attribute_type,
                            "required": True,
                            "default": default_json,
                            "extensions": {"color": None, "description": None},
                        },
                    )
                )

        return cast(
            JsonObject,
            {
                "type": provider,
                "slug": slug,
                "integration_name": integration_name,
                "setup_attributes": setup_attributes,
            },
        )

    @staticmethod
    def _model_dump(value: Any) -> Any:
        if hasattr(value, "model_dump"):
            return value.model_dump(mode="json")
        if hasattr(value, "dict"):
            return value.dict()
        return value

    def _apply_auto_layout(self, nodes_dict_list: List[Dict[str, Any]], edges: List[Json]) -> None:
        """Apply auto-layout to nodes that are all positioned at (0,0)."""
        nodes_for_layout: List[Tuple[str, NodeDisplayData]] = []
        for node_dict in nodes_dict_list:
            if isinstance(node_dict.get("id"), str) and isinstance(node_dict.get("display_data"), dict):
                display_data = node_dict["display_data"]
                position = display_data.get("position", {})
                if isinstance(position, dict):
                    nodes_for_layout.append(
                        (
                            str(node_dict["id"]),
                            NodeDisplayData(
                                position=NodeDisplayPosition(
                                    x=float(position.get("x", 0.0)), y=float(position.get("y", 0.0))
                                ),
                                width=display_data.get("width"),
                                height=display_data.get("height"),
                                comment=display_data.get("comment"),
                            ),
                        )
                    )

        edges_for_layout: List[Tuple[str, str, EdgeDisplay]] = []
        for edge in edges:
            if isinstance(edge, dict):
                edge_dict = cast(Dict[str, Any], edge)
                edge_source_node_id: Optional[Any] = edge_dict.get("source_node_id")
                edge_target_node_id: Optional[Any] = edge_dict.get("target_node_id")
                edge_id_raw: Optional[Any] = edge_dict.get("id")
                if (
                    isinstance(edge_source_node_id, str)
                    and isinstance(edge_target_node_id, str)
                    and isinstance(edge_id_raw, str)
                ):
                    edges_for_layout.append(
                        (edge_source_node_id, edge_target_node_id, EdgeDisplay(id=UUID(edge_id_raw)))
                    )

        positioned_nodes = auto_layout_nodes(nodes_for_layout, edges_for_layout)

        for node_id, positioned_data in positioned_nodes:
            for node_dict in nodes_dict_list:
                node_id_val = node_dict.get("id")
                display_data = node_dict.get("display_data")
                if isinstance(node_id_val, str) and node_id_val == node_id and isinstance(display_data, dict):
                    display_data_dict = cast(Dict[str, Any], display_data)
                    display_data_dict["position"] = positioned_data.position.dict()

    @cached_property
    def workflow_id(self) -> UUID:
        """Can be overridden as a class attribute to specify a custom workflow id."""
        return self._workflow.__id__

    def _enrich_global_node_output_displays(
        self,
        node: Type[BaseNode],
        node_display: BaseNodeDisplay,
        node_output_displays: Dict[OutputReference, NodeOutputDisplay],
    ):
        """This method recursively adds nodes wrapped in decorators to the node_output_displays dictionary."""

        inner_node = get_wrapped_node(node)
        if inner_node:
            inner_node_display = self._get_node_display(inner_node)
            self._enrich_global_node_output_displays(inner_node, inner_node_display, node_output_displays)

        for node_output in node.Outputs:
            if node_output in node_output_displays:
                continue

            node_output_displays[node_output] = node_display.get_node_output_display(node_output)

    def _enrich_node_port_displays(
        self,
        node: Type[BaseNode],
        node_display: BaseNodeDisplay,
        port_displays: Dict[Port, PortDisplay],
    ):
        """This method recursively adds nodes wrapped in decorators to the port_displays dictionary."""

        inner_node = get_wrapped_node(node)
        if inner_node:
            inner_node_display = self._get_node_display(inner_node)
            self._enrich_node_port_displays(inner_node, inner_node_display, port_displays)

        for port in node.Ports:
            if port in port_displays:
                continue

            port_displays[port] = node_display.get_node_port_display(port)

    def _get_node_display(self, node: Type[BaseNode]) -> BaseNodeDisplay:
        node_display_class = get_node_display_class(node)
        return node_display_class()

    @cached_property
    def display_context(self) -> WorkflowDisplayContext:
        workflow_meta_display = self._generate_workflow_meta_display()

        global_node_output_displays: NodeOutputDisplays = (
            copy(self._parent_display_context.global_node_output_displays) if self._parent_display_context else {}
        )

        node_displays: NodeDisplays = {}

        global_node_displays: NodeDisplays = (
            copy(self._parent_display_context.global_node_displays) if self._parent_display_context else {}
        )

        port_displays: PortDisplays = {}

        for node in self._workflow.get_all_nodes():
            self._enrich_node_displays(
                node=node,
                node_displays=node_displays,
                global_node_displays=global_node_displays,
                global_node_output_displays=global_node_output_displays,
                port_displays=port_displays,
            )

        workflow_input_displays: WorkflowInputsDisplays = {}
        # If we're dealing with a nested workflow, then it should have access to the inputs of its parents.
        global_workflow_input_displays = (
            copy(self._parent_display_context.global_workflow_input_displays) if self._parent_display_context else {}
        )
        for workflow_input in self._workflow.get_inputs_class():
            workflow_input_display_overrides = self.inputs_display.get(workflow_input)
            input_display = self._generate_workflow_input_display(
                workflow_input, overrides=workflow_input_display_overrides
            )
            workflow_input_displays[workflow_input] = input_display
            global_workflow_input_displays[workflow_input] = input_display

        state_value_displays: StateValueDisplays = {}
        global_state_value_displays = (
            copy(self._parent_display_context.global_state_value_displays) if self._parent_display_context else {}
        )
        for state_value in self._workflow.get_state_class():
            state_value_display_overrides = self.state_value_displays.get(state_value)
            state_value_display = self._generate_state_value_display(
                state_value, overrides=state_value_display_overrides
            )
            state_value_displays[state_value] = state_value_display
            global_state_value_displays[state_value] = state_value_display

        entrypoint_displays: EntrypointDisplays = {}
        for entrypoint in self._workflow.get_entrypoints():
            if entrypoint in entrypoint_displays:
                continue

            entrypoint_display_overrides = self.entrypoint_displays.get(entrypoint)
            entrypoint_displays[entrypoint] = self._generate_entrypoint_display(
                entrypoint, workflow_meta_display, node_displays, overrides=entrypoint_display_overrides
            )

        edge_displays: Dict[Tuple[Port, Type[BaseNode]], EdgeDisplay] = {}
        for edge in self._workflow.get_edges():
            if edge in edge_displays:
                continue

            edge_display_overrides = self.edge_displays.get((edge.from_port, edge.to_node))
            edge_displays[(edge.from_port, edge.to_node)] = edge_display_overrides or self._generate_edge_display(
                edge, node_displays
            )

        for edge in self._workflow.get_unused_edges():
            if edge in edge_displays:
                continue

            edge_display_overrides = self.edge_displays.get((edge.from_port, edge.to_node))
            edge_displays[(edge.from_port, edge.to_node)] = edge_display_overrides or self._generate_edge_display(
                edge, node_displays
            )

        workflow_output_displays: Dict[BaseDescriptor, WorkflowOutputDisplay] = {}
        for workflow_output in self._workflow.Outputs:
            if workflow_output in workflow_output_displays:
                continue

            if not isinstance(workflow_output, OutputReference):
                raise ValueError(f"{workflow_output} must be an {OutputReference.__name__}")

            workflow_output_display = self.output_displays.get(workflow_output)
            workflow_output_displays[workflow_output] = (
                workflow_output_display or self._generate_workflow_output_display(workflow_output)
            )

        return WorkflowDisplayContext(
            client=self._client,
            workflow_display=workflow_meta_display,
            workflow_input_displays=workflow_input_displays,
            global_workflow_input_displays=global_workflow_input_displays,
            state_value_displays=state_value_displays,
            global_state_value_displays=global_state_value_displays,
            node_displays=node_displays,
            global_node_output_displays=global_node_output_displays,
            global_node_displays=global_node_displays,
            entrypoint_displays=entrypoint_displays,
            workflow_output_displays=workflow_output_displays,
            edge_displays=edge_displays,
            port_displays=port_displays,
            workflow_display_class=self.__class__,
            dry_run=self._dry_run,
        )

    def _generate_workflow_meta_display(self) -> WorkflowMetaDisplay:
        overrides = self.workflow_display
        if overrides:
            return WorkflowMetaDisplay(
                entrypoint_node_id=overrides.entrypoint_node_id,
                entrypoint_node_source_handle_id=overrides.entrypoint_node_source_handle_id,
                entrypoint_node_display=overrides.entrypoint_node_display,
                display_data=overrides.display_data,
            )

        return WorkflowMetaDisplay.get_default(self._workflow)

    def _generate_workflow_input_display(
        self, workflow_input: WorkflowInputReference, overrides: Optional[WorkflowInputsDisplay] = None
    ) -> WorkflowInputsDisplay:
        workflow_input_id: UUID
        name = None
        color = None
        if overrides:
            workflow_input_id = overrides.id
            name = overrides.name
            color = overrides.color
        else:
            workflow_input_id = workflow_input.id

        return WorkflowInputsDisplay(id=workflow_input_id, name=name, color=color)

    def _generate_state_value_display(
        self, state_value: StateValueReference, overrides: Optional[StateValueDisplay] = None
    ) -> StateValueDisplay:
        state_value_id: UUID
        name = None
        color = None
        if overrides:
            state_value_id = overrides.id
            name = overrides.name
            color = overrides.color
        else:
            state_value_id = state_value.id

        return StateValueDisplay(id=state_value_id, name=name, color=color)

    def _generate_entrypoint_display(
        self,
        entrypoint: Type[BaseNode],
        workflow_display: WorkflowMetaDisplay,
        node_displays: Dict[Type[BaseNode], BaseNodeDisplay],
        overrides: Optional[EntrypointDisplay] = None,
    ) -> EntrypointDisplay:
        entrypoint_node_id = workflow_display.entrypoint_node_id

        edge_display_overrides = overrides.edge_display if overrides else None
        entrypoint_id = (
            edge_display_overrides.id
            if edge_display_overrides
            else uuid4_from_hash(f"{self.workflow_id}|id|{entrypoint_node_id}")
        )

        entrypoint_target = get_unadorned_node(entrypoint)
        target_node_display = node_displays[entrypoint_target]
        target_node_id = target_node_display.node_id

        edge_display = edge_display_overrides or self._generate_edge_display_from_source(
            entrypoint_node_id, target_node_id
        )

        return EntrypointDisplay(id=entrypoint_id, edge_display=edge_display)

    def _generate_workflow_output_display(self, output: OutputReference) -> WorkflowOutputDisplay:
        return WorkflowOutputDisplay(id=output.id, name=output.name)

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)

        workflow_class = get_args(cls.__orig_bases__[0])[0]  # type: ignore [attr-defined]
        register_workflow_display_class(workflow_class=workflow_class, workflow_display_class=cls)

    @staticmethod
    def gather_event_display_context(
        module_path: str,
        # DEPRECATED: This will be removed in the 0.15.0 release
        workflow_class: Optional[Type[BaseWorkflow]] = None,
    ) -> Union[WorkflowEventDisplayContext, None]:
        full_workflow_display_module_path = f"{module_path}.display.workflow"
        try:
            display_module = importlib.import_module(full_workflow_display_module_path)
        except ModuleNotFoundError:
            return BaseWorkflowDisplay._gather_event_display_context_from_workflow_crawling(module_path, workflow_class)

        WorkflowDisplayClass: Optional[Type[BaseWorkflowDisplay]] = None
        for name, definition in display_module.__dict__.items():
            if name.startswith("_"):
                continue

            if (
                not isinstance(definition, type)
                or not issubclass(definition, BaseWorkflowDisplay)
                or definition == BaseWorkflowDisplay
            ):
                continue

            WorkflowDisplayClass = definition
            break

        if WorkflowDisplayClass:
            return WorkflowDisplayClass().get_event_display_context()

        return BaseWorkflowDisplay._gather_event_display_context_from_workflow_crawling(module_path, workflow_class)

    @staticmethod
    def _gather_event_display_context_from_workflow_crawling(
        module_path: str,
        workflow_class: Optional[Type[BaseWorkflow]] = None,
    ) -> Union[WorkflowEventDisplayContext, None]:
        try:
            if workflow_class is None:
                workflow_class = BaseWorkflow.load_from_module(module_path)

            workflow_display = get_workflow_display(workflow_class=workflow_class)
            return workflow_display.get_event_display_context()

        except ModuleNotFoundError:
            logger.exception("Failed to load workflow from module %s", module_path)
            return None

    def get_event_display_context(self):
        display_context = self.display_context

        workflow_outputs = {
            output.name: display_context.workflow_output_displays[output].id
            for output in display_context.workflow_output_displays
        }
        workflow_inputs = {
            input.name: display_context.workflow_input_displays[input].id
            for input in display_context.workflow_input_displays
        }
        node_displays = {
            node.__id__: (node, display_context.node_displays[node]) for node in display_context.node_displays
        }
        node_event_displays = {}
        for node_id in node_displays:
            node, current_node_display = node_displays[node_id]
            input_display = current_node_display.node_input_ids_by_name
            output_display = {
                output.name: current_node_display.output_display[output].id
                for output in current_node_display.output_display
            }
            port_display_meta = {
                port.name: current_node_display.port_displays[port].id for port in current_node_display.port_displays
            }
            subworkflow_display_context: Optional[WorkflowEventDisplayContext] = None
            if hasattr(node, "subworkflow"):
                # All nodes that have a subworkflow attribute are currently expected to call them `subworkflow`
                # This will change if in the future we decide to support multiple subworkflows on a single node
                subworkflow_attribute = raise_if_descriptor(getattr(node, "subworkflow"))
                if issubclass(subworkflow_attribute, BaseWorkflow):
                    subworkflow_display = get_workflow_display(
                        base_display_class=display_context.workflow_display_class,
                        workflow_class=subworkflow_attribute,
                        parent_display_context=display_context,
                    )
                    subworkflow_display_context = subworkflow_display.get_event_display_context()

            node_event_displays[node_id] = NodeEventDisplayContext(
                input_display=input_display,
                output_display=output_display,
                port_display=port_display_meta,
                subworkflow_display=subworkflow_display_context,
            )

        display_meta = WorkflowEventDisplayContext(
            workflow_outputs=workflow_outputs,
            workflow_inputs=workflow_inputs,
            node_displays=node_event_displays,
        )
        return display_meta

    def _enrich_node_displays(
        self,
        node: Type[BaseNode],
        node_displays: NodeDisplays,
        global_node_displays: NodeDisplays,
        global_node_output_displays: NodeOutputDisplays,
        port_displays: PortDisplays,
    ) -> None:
        extracted_node_displays = self._extract_node_displays(node)

        for extracted_node, extracted_node_display in extracted_node_displays.items():
            if extracted_node not in node_displays:
                node_displays[extracted_node] = extracted_node_display

            if extracted_node not in global_node_displays:
                global_node_displays[extracted_node] = extracted_node_display

        self._enrich_global_node_output_displays(node, extracted_node_displays[node], global_node_output_displays)
        self._enrich_node_port_displays(node, extracted_node_displays[node], port_displays)

    def _extract_node_displays(self, node: Type[BaseNode]) -> Dict[Type[BaseNode], BaseNodeDisplay]:
        node_display = self._get_node_display(node)
        additional_node_displays: Dict[Type[BaseNode], BaseNodeDisplay] = {
            node: node_display,
        }

        # Nodes wrapped in a decorator need to be in our node display dictionary for later retrieval
        inner_node = get_wrapped_node(node)
        if inner_node:
            inner_node_displays = self._extract_node_displays(inner_node)

            for node, display in inner_node_displays.items():
                if node not in additional_node_displays:
                    additional_node_displays[node] = display

        return additional_node_displays

    def _generate_edge_display(self, edge: Edge, node_displays: Dict[Type[BaseNode], BaseNodeDisplay]) -> EdgeDisplay:
        source_node = get_unadorned_node(edge.from_port.node_class)
        target_node = get_unadorned_node(edge.to_node)

        source_node_id = node_displays[source_node].node_id
        target_node_id = node_displays[target_node].node_id

        return self._generate_edge_display_from_source(source_node_id, target_node_id)

    def _generate_edge_display_from_source(
        self,
        source_node_id: UUID,
        target_node_id: UUID,
    ) -> EdgeDisplay:
        return EdgeDisplay(
            id=uuid4_from_hash(f"{self.workflow_id}|id|{source_node_id}|{target_node_id}"),
        )

    @classmethod
    def infer_workflow_class(cls) -> Type[BaseWorkflow]:
        original_base = get_original_base(cls)
        workflow_class = get_args(original_base)[0]
        if isinstance(workflow_class, TypeVar):
            bounded_class = workflow_class.__bound__
            if inspect.isclass(bounded_class) and issubclass(bounded_class, BaseWorkflow):
                return bounded_class

            if isinstance(bounded_class, ForwardRef) and bounded_class.__forward_arg__ == BaseWorkflow.__name__:
                return BaseWorkflow

        if issubclass(workflow_class, BaseWorkflow):
            return workflow_class

        raise ValueError(f"Workflow {cls.__name__} must be a subclass of {BaseWorkflow.__name__}")

    @property
    def _workflow(self) -> Type[WorkflowType]:
        return cast(Type[WorkflowType], self.__class__.infer_workflow_class())

    @staticmethod
    def serialize_module(
        module: str,
        *,
        client: Optional[VellumClient] = None,
        dry_run: bool = False,
    ) -> WorkflowSerializationResult:
        """
        Load a workflow from a module and serialize it to JSON.

        Args:
            module: The module path to load the workflow from
            client: Optional Vellum client to use for serialization
            dry_run: Whether to run in dry-run mode

        Returns:
            WorkflowSerializationResult containing exec_config and errors
        """
        workflow = BaseWorkflow.load_from_module(module)
        workflow_display = get_workflow_display(
            workflow_class=workflow,
            client=client,
            dry_run=dry_run,
        )

        exec_config = workflow_display.serialize()
        additional_files = workflow_display._gather_additional_module_files(module)

        if additional_files:
            exec_config["module_data"] = {"additional_files": cast(JsonObject, additional_files)}

        dataset = None
        try:
            sandbox_module_path = f"{module}.sandbox"
            sandbox_module = importlib.import_module(sandbox_module_path)
            if hasattr(sandbox_module, "dataset"):
                dataset_attr = getattr(sandbox_module, "dataset")
                if dataset_attr and isinstance(dataset_attr, list):
                    dataset = []
                    dataset_row_index_to_id = load_dataset_row_index_to_id_mapping(module)
                    for i, inputs_obj in enumerate(dataset_attr):
                        normalized_row = (
                            DatasetRow(label=f"Scenario {i + 1}", inputs=inputs_obj)
                            if isinstance(inputs_obj, BaseInputs)
                            else inputs_obj
                        )
                        row_data = normalized_row.model_dump(
                            mode="json",
                            by_alias=True,
                            exclude_none=True,
                            context={
                                "serializer": lambda value: serialize_value(
                                    workflow_display.workflow_id,
                                    workflow_display.display_context,
                                    value,
                                )
                            },
                        )

                        if i in dataset_row_index_to_id:
                            row_data["id"] = dataset_row_index_to_id[i]
                        elif isinstance(inputs_obj, DatasetRow) and inputs_obj.id is not None:
                            row_data["id"] = inputs_obj.id

                        dataset.append(row_data)
        except (ImportError, AttributeError):
            pass

        return WorkflowSerializationResult(
            exec_config=exec_config,
            errors=[
                WorkflowSerializationError(
                    message=str(error),
                    stacktrace="".join(traceback.format_exception(type(error), error, error.__traceback__)),
                )
                for error in workflow_display.display_context.errors
            ],
            dataset=dataset,
        )

    def _gather_additional_module_files(self, module_path: str) -> Dict[str, str]:
        workflow_module_path = f"{module_path}.workflow"
        workflow_module = importlib.import_module(workflow_module_path)

        workflow_file_path = workflow_module.__file__
        if not workflow_file_path:
            return {}

        module_dir = os.path.dirname(workflow_file_path)
        additional_files = {}

        for root, _, filenames in os.walk(module_dir):
            for filename in filenames:
                file_path = os.path.join(root, filename)
                relative_path = os.path.relpath(file_path, start=module_dir)

                should_ignore = False
                for ignore_pattern in IGNORE_PATTERNS:
                    if fnmatch.fnmatch(filename, ignore_pattern) or fnmatch.fnmatch(relative_path, ignore_pattern):
                        should_ignore = True
                        break

                if not should_ignore:
                    for serialized_pattern in self._serialized_files:
                        if "*" in serialized_pattern:
                            if fnmatch.fnmatch(relative_path, serialized_pattern) or fnmatch.fnmatch(
                                filename, serialized_pattern
                            ):
                                should_ignore = True
                                break
                        else:
                            if relative_path == serialized_pattern:
                                should_ignore = True
                                break

                if should_ignore:
                    continue

                try:
                    with open(file_path, encoding="utf-8") as f:
                        additional_files[relative_path] = f.read()
                except (UnicodeDecodeError, PermissionError):
                    continue

        return additional_files

    @staticmethod
    def _is_reference_required(reference: BaseDescriptor) -> bool:
        has_default = reference.instance is not undefined
        is_optional = type(None) in reference.types
        is_required = not has_default and not is_optional
        return is_required

    def _is_node_invalid(self, node: Type[BaseNode]) -> bool:
        """Check if a node failed to serialize and should be considered invalid."""
        return node in self.display_context.invalid_nodes


register_workflow_display_class(workflow_class=BaseWorkflow, workflow_display_class=BaseWorkflowDisplay)



import pytest
from uuid import UUID
from typing import Type

from vellum.client.core.api_error import ApiError
from vellum.workflows.environment import EnvironmentVariables
from vellum.workflows.nodes.displayable.code_execution_node.node import CodeExecutionNode
from vellum.workflows.references.vellum_secret import VellumSecretReference
from vellum.workflows.state.base import BaseState
from vellum.workflows.workflows.base import BaseWorkflow
from vellum_ee.workflows.display.nodes.vellum.code_execution_node import BaseCodeExecutionNodeDisplay
from vellum_ee.workflows.display.utils.exceptions import NodeValidationError
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


def _no_display_class(Node: Type[CodeExecutionNode]):
    return None


def _display_class_with_node_input_ids_by_name(Node: Type[CodeExecutionNode]):
    class CodeExecutionNodeDisplay(BaseCodeExecutionNodeDisplay[Node]):  # type: ignore[valid-type]
        node_input_ids_by_name = {"foo": UUID("fba6a4d5-835a-4e99-afb7-f6a4aed15110")}

    return CodeExecutionNodeDisplay


def _display_class_with_node_input_ids_by_name_with_inputs_prefix(Node: Type[CodeExecutionNode]):
    class CodeExecutionNodeDisplay(BaseCodeExecutionNodeDisplay[Node]):  # type: ignore[valid-type]
        node_input_ids_by_name = {"code_inputs.foo": UUID("fba6a4d5-835a-4e99-afb7-f6a4aed15110")}

    return CodeExecutionNodeDisplay


@pytest.mark.parametrize(
    ["GetDisplayClass", "expected_input_id"],
    [
        (_no_display_class, "20ff166f-af59-4515-8ff5-205226c01aa4"),
        (_display_class_with_node_input_ids_by_name, "fba6a4d5-835a-4e99-afb7-f6a4aed15110"),
        (_display_class_with_node_input_ids_by_name_with_inputs_prefix, "fba6a4d5-835a-4e99-afb7-f6a4aed15110"),
    ],
    ids=[
        "no_display_class",
        "display_class_with_node_input_ids_by_name",
        "display_class_with_node_input_ids_by_name_with_inputs_prefix",
    ],
)
def test_serialize_node__code_node_inputs(GetDisplayClass, expected_input_id):
    # GIVEN a code node with inputs
    class MyCodeExecutionNode(CodeExecutionNode):
        code_inputs = {"foo": "bar"}

    # AND a workflow with the code node
    class Workflow(BaseWorkflow):
        graph = MyCodeExecutionNode

    # AND a display class
    GetDisplayClass(MyCodeExecutionNode)

    # WHEN the workflow is serialized
    workflow_display = get_workflow_display(workflow_class=Workflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN the node should properly serialize the inputs
    my_code_execution_node = next(
        node for node in serialized_workflow["workflow_raw_data"]["nodes"] if node["type"] == "CODE_EXECUTION"
    )

    assert my_code_execution_node["inputs"] == [
        {
            "id": expected_input_id,
            "key": "foo",
            "value": {
                "combinator": "OR",
                "rules": [
                    {
                        "type": "CONSTANT_VALUE",
                        "data": {
                            "type": "STRING",
                            "value": "bar",
                        },
                    }
                ],
            },
        },
        {
            "id": "50678b9f-bdea-41c0-bc3d-425ea38466ee",
            "key": "code",
            "value": {
                "combinator": "OR",
                "rules": [
                    {
                        "type": "CONSTANT_VALUE",
                        "data": {
                            "type": "STRING",
                            "value": "",
                        },
                    }
                ],
            },
        },
        {
            "id": "6d2840f3-a5c1-4376-8616-ced4fffc6cf2",
            "key": "runtime",
            "value": {
                "combinator": "OR",
                "rules": [
                    {
                        "type": "CONSTANT_VALUE",
                        "data": {
                            "type": "STRING",
                            "value": "PYTHON_3_11_6",
                        },
                    }
                ],
            },
        },
    ]


def test_serialize_node__with_unresolved_secret_references(vellum_client):
    # GIVEN a node has access to a secret reference
    class MyNode(CodeExecutionNode):
        code_inputs = {"api_key": VellumSecretReference("MY_API_KEY")}

    # AND the secret is not found
    vellum_client.workspace_secrets.retrieve.side_effect = ApiError(
        status_code=404,
        body="Secret not found",
    )

    # AND a workflow with the code node
    class Workflow(BaseWorkflow):
        graph = MyNode

    # WHEN we serialize the workflow
    workflow_display = get_workflow_display(workflow_class=Workflow)
    data: dict = workflow_display.serialize()

    # THEN the condition should be serialized correctly
    node = next(node for node in data["workflow_raw_data"]["nodes"] if node["type"] == "CODE_EXECUTION")
    assert node["inputs"][0]["value"] == {
        "combinator": "OR",
        "rules": [
            {
                "type": "WORKSPACE_SECRET",
                "data": {
                    "type": "STRING",
                    "workspace_secret_id": None,
                },
            }
        ],
    }

    # AND we should have a warning of the invalid reference
    # TODO: Come up with a proposal for how nodes should propagate warnings
    # warnings = list(workflow_display.errors)
    # assert len(warnings) == 1
    # assert "Failed to resolve secret reference 'MY_API_KEY'" in str(warnings[0])


def test_serialize_node__with_non_exist_code_input_path():
    # GIVEN a code node with a non-existent code input path
    class MyNode(CodeExecutionNode):
        filepath = "non_existent_file.py"

    # AND a workflow with the code node
    class Workflow(BaseWorkflow):
        graph = MyNode

    # WHEN we serialize the workflow
    workflow_display = get_workflow_display(workflow_class=Workflow)
    with pytest.raises(NodeValidationError) as exc_info:
        workflow_display.serialize()
    assert "Filepath 'non_existent_file.py' does not exist" in str(exc_info.value)


def test_serialize_node__with_environment_variable_references():
    """
    Tests that environment variable references in code node inputs serialize correctly.
    """

    # GIVEN a code node with environment variable references in code_inputs
    class MyCodeExecutionNode(CodeExecutionNode):
        code_inputs = {
            "api_key": EnvironmentVariables.get("MY_API_KEY"),
            "other_config": {"nested_key": EnvironmentVariables.get("NESTED_KEY")},
        }

    # AND a workflow with the code node
    class Workflow(BaseWorkflow):
        graph = MyCodeExecutionNode

    # WHEN the workflow is serialized
    workflow_display = get_workflow_display(workflow_class=Workflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN the node should properly serialize the environment variable references
    my_code_execution_node = next(
        node for node in serialized_workflow["workflow_raw_data"]["nodes"] if node["type"] == "CODE_EXECUTION"
    )

    # AND the api_key input should be serialized as an ENVIRONMENT_VARIABLE
    api_key_input = next(inp for inp in my_code_execution_node["inputs"] if inp["key"] == "api_key")
    assert api_key_input["value"] == {
        "combinator": "OR",
        "rules": [
            {
                "type": "ENVIRONMENT_VARIABLE",
                "data": {
                    "environment_variable": "MY_API_KEY",
                },
            }
        ],
    }

    # AND the nested environment variable should also be serialized correctly
    other_config_input = next(inp for inp in my_code_execution_node["inputs"] if inp["key"] == "other_config")
    assert other_config_input["value"]["combinator"] == "OR"
    assert len(other_config_input["value"]["rules"]) == 1
    assert other_config_input["value"]["rules"][0]["type"] == "CONSTANT_VALUE"
    # The nested dict should contain the environment variable reference
    nested_data = other_config_input["value"]["rules"][0]["data"]
    assert nested_data["type"] == "JSON"
    assert "nested_key" in nested_data["value"]


def test_serialize_node__with_non_exist_code_input_path_with_dry_run():
    # GIVEN a code node with a non-existent code input path
    class MyNode(CodeExecutionNode):
        filepath = "non_existent_file.py"

    # AND a workflow with the code node
    class Workflow(BaseWorkflow):
        graph = MyNode

    # WHEN we serialize the workflow with dry_run=True
    workflow_display = get_workflow_display(workflow_class=Workflow, dry_run=True)
    data: dict = workflow_display.serialize()

    # THEN the workflow should not raise an error
    assert data == {
        "workflow_raw_data": {
            "nodes": [
                {
                    "id": "9b9e2a5d-01a4-46b2-80a3-d9484b2c0e08",
                    "type": "ENTRYPOINT",
                    "inputs": [],
                    "data": {"label": "Entrypoint Node", "source_handle_id": "3e2a3f52-5047-4e2e-9a21-37bd43c63250"},
                    "display_data": {"position": {"x": 0.0, "y": -50.0}},
                    "base": None,
                    "definition": None,
                },
                {
                    "id": "f41cebba-a048-4852-a3d7-0f3100927166",
                    "type": "CODE_EXECUTION",
                    "inputs": [
                        {
                            "id": "b807bd5f-7d49-4597-89cf-044aa84cf8d6",
                            "key": "code",
                            "value": {
                                "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "JSON", "value": None}}],
                                "combinator": "OR",
                            },
                        },
                        {
                            "id": "dcd1e1a4-49e2-4d6b-95f9-8ec5fd2f8f5d",
                            "key": "runtime",
                            "value": {
                                "rules": [
                                    {"type": "CONSTANT_VALUE", "data": {"type": "STRING", "value": "PYTHON_3_11_6"}}
                                ],
                                "combinator": "OR",
                            },
                        },
                    ],
                    "data": {
                        "label": "My Node",
                        "error_output_id": None,
                        "source_handle_id": "dc9edb2e-4392-4a2c-ab92-cc1b9c0cbd53",
                        "target_handle_id": "66e7ef63-518b-40e7-911a-e38e8bcaec81",
                        "code_input_id": "b807bd5f-7d49-4597-89cf-044aa84cf8d6",
                        "runtime_input_id": "dcd1e1a4-49e2-4d6b-95f9-8ec5fd2f8f5d",
                        "output_type": "STRING",
                        "packages": [],
                        "output_id": "98cae9b9-45cc-4897-a0f5-df250b56c00d",
                        "log_output_id": "66c06c97-a9d1-4abf-840f-3f6c29709612",
                    },
                    "display_data": {"position": {"x": 200.0, "y": -50.0}},
                    "base": {
                        "name": "CodeExecutionNode",
                        "module": ["vellum", "workflows", "nodes", "displayable", "code_execution_node", "node"],
                    },
                    "definition": {
                        "name": "MyNode",
                        "module": [
                            "vellum_ee",
                            "workflows",
                            "display",
                            "nodes",
                            "vellum",
                            "tests",
                            "test_code_execution_node",
                        ],
                    },
                    "ports": [{"id": "dc9edb2e-4392-4a2c-ab92-cc1b9c0cbd53", "name": "default", "type": "DEFAULT"}],
                    "trigger": {"id": "66e7ef63-518b-40e7-911a-e38e8bcaec81", "merge_behavior": "AWAIT_ANY"},
                    "outputs": [
                        {
                            "id": "98cae9b9-45cc-4897-a0f5-df250b56c00d",
                            "name": "result",
                            "schema": {"type": "string"},
                            "type": "STRING",
                            "value": None,
                        },
                        {
                            "id": "66c06c97-a9d1-4abf-840f-3f6c29709612",
                            "name": "log",
                            "schema": {"type": "string"},
                            "type": "STRING",
                            "value": None,
                        },
                    ],
                },
            ],
            "edges": [
                {
                    "id": "85e0961e-f968-49a0-beed-e21373f0ecda",
                    "source_node_id": "9b9e2a5d-01a4-46b2-80a3-d9484b2c0e08",
                    "source_handle_id": "3e2a3f52-5047-4e2e-9a21-37bd43c63250",
                    "target_node_id": "f41cebba-a048-4852-a3d7-0f3100927166",
                    "target_handle_id": "66e7ef63-518b-40e7-911a-e38e8bcaec81",
                    "type": "DEFAULT",
                }
            ],
            "display_data": {"viewport": {"x": 0.0, "y": 0.0, "zoom": 1.0}},
            "definition": {
                "name": "Workflow",
                "module": ["vellum_ee", "workflows", "display", "nodes", "vellum", "tests", "test_code_execution_node"],
            },
            "output_values": [],
        },
        "input_variables": [],
        "state_variables": [],
        "output_variables": [],
    }


def test_serialize_node__with_custom_output_type():
    # GIVEN a code node with a custom output type
    class MyNode(CodeExecutionNode[BaseState, dict[str, int]]):
        code = """\
return {
    "hello": 1,
}
"""

    # AND a workflow with the code node
    class Workflow(BaseWorkflow):
        graph = MyNode

    # WHEN we serialize the workflow
    workflow_display = get_workflow_display(workflow_class=Workflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN the node's outputs should serialize correctly
    my_code_execution_node = next(
        node for node in serialized_workflow["workflow_raw_data"]["nodes"] if node["type"] == "CODE_EXECUTION"
    )
    assert my_code_execution_node["outputs"] == [
        {
            "id": "01de8e8b-5e0e-4344-93b0-e002bbaed840",
            "name": "result",
            "value": None,
            "type": "JSON",
            "schema": {"type": "object", "additionalProperties": {"type": "integer"}},
        },
        {
            "id": "64bf62e0-adc7-48cc-b689-8bf9b7e4eeef",
            "name": "log",
            "value": None,
            "type": "STRING",
            "schema": {"type": "string"},
        },
    ]
    assert my_code_execution_node["data"]["output_type"] == "JSON"



from vellum.workflows import BaseInputs, BaseNode, BaseState, BaseWorkflow, MockNodeExecution
from vellum_ee.workflows.display.utils.expressions import base_descriptor_validator


def test_mocks__parse_from_app__descriptors():
    # GIVEN a Base Node
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    # AND workflow inputs
    class Inputs(BaseInputs):
        bar: str

    # AND workflow state
    class State(BaseState):
        baz: str

    # AND a workflow class with that Node
    class MyWorkflow(BaseWorkflow[Inputs, State]):
        graph = StartNode

        class Outputs(BaseWorkflow.Outputs):
            final_value = StartNode.Outputs.foo

    # AND a mock workflow node execution from the app
    raw_mock_workflow_node_executions = [
        {
            "node_id": str(StartNode.__id__),
            "when_condition": {
                "type": "BINARY_EXPRESSION",
                "operator": ">=",
                "lhs": {
                    "type": "EXECUTION_COUNTER",
                    "node_id": str(StartNode.__id__),
                },
                "rhs": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "NUMBER",
                        "value": 1,
                    },
                },
            },
            "then_outputs": {
                "foo": "Hello foo",
            },
        },
        {
            "node_id": str(StartNode.__id__),
            "when_condition": {
                "type": "BINARY_EXPRESSION",
                "operator": "==",
                "lhs": {
                    "type": "WORKFLOW_INPUT",
                    "input_variable_id": str(Inputs.bar.id),
                },
                "rhs": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "STRING",
                        "value": "bar",
                    },
                },
            },
            "then_outputs": {
                "foo": "Hello bar",
            },
        },
        {
            "node_id": str(StartNode.__id__),
            "when_condition": {
                "type": "BINARY_EXPRESSION",
                "operator": "!=",
                "lhs": {
                    "type": "WORKFLOW_STATE",
                    "state_variable_id": str(State.baz.id),
                },
                "rhs": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "STRING",
                        "value": "baz",
                    },
                },
            },
            "then_outputs": {
                "foo": "Hello baz",
            },
        },
    ]

    # WHEN we parsed the raw data on `MockNodeExecution`
    node_output_mocks = MockNodeExecution.validate_all(
        raw_mock_workflow_node_executions,
        MyWorkflow,
        descriptor_validator=base_descriptor_validator,
    )

    # THEN we get the expected list of MockNodeExecution objects
    assert node_output_mocks
    assert len(node_output_mocks) == 3
    assert node_output_mocks[0] == MockNodeExecution(
        when_condition=StartNode.Execution.count.greater_than_or_equal_to(1),
        then_outputs=StartNode.Outputs(
            foo="Hello foo",
        ),
    )
    assert node_output_mocks[1] == MockNodeExecution(
        when_condition=Inputs.bar.equals("bar"),
        then_outputs=StartNode.Outputs(
            foo="Hello bar",
        ),
    )
    assert node_output_mocks[2] == MockNodeExecution(
        when_condition=State.baz.does_not_equal("baz"),
        then_outputs=StartNode.Outputs(
            foo="Hello baz",
        ),
    )



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .workspace_display_config import WorkspaceDisplayConfig


class WorkspaceRead(UniversalBaseModel):
    id: str
    name: str = pydantic.Field()
    """
    The name of the Workspace.
    """

    label: str
    created: dt.datetime
    display_config: typing.Optional[WorkspaceDisplayConfig] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel


class VellumImageRequest(UniversalBaseModel):
    src: str = pydantic.Field()
    """
    The reference to the source data. This can be one of the following formats:
    1. A base64-encoded data URL (e.g., 'data:application/pdf;base64,JVBERi0xLjQKJcfs...').
    2. A url pointing to a file accessible over HTTP or HTTPS.
    3. A reference to a file that's been previously uploaded to Vellum, in the form of 'vellum:uploaded-file:{uploaded_file_id}'.
        You can use the Uploaded Files API to retrieve the url of the uploaded file given its ID. See https://docs.vellum.ai/developers/client-sdk/uploaded-files/retrieve for details.
    """

    metadata: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import pytest

from vellum.workflows import BaseWorkflow
from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.ports.port import Port
from vellum.workflows.state.base import BaseState
from vellum_ee.workflows.display.utils.exceptions import WorkflowValidationError
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


def test_workflow_serialization_error__node_points_to_itself():
    """
    Tests that serialization raises an error when a node creates an infinite loop by pointing to itself.
    """

    # GIVEN a simple workflow where a node points to itself
    class StartNode(BaseNode[BaseState]):
        pass

    class InfiniteLoopWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = StartNode >> StartNode

    # WHEN we attempt to serialize the workflow
    workflow_display = get_workflow_display(workflow_class=InfiniteLoopWorkflow)

    # THEN it should raise a WorkflowValidationError about the self-edge
    with pytest.raises(WorkflowValidationError) as exc_info:
        workflow_display.serialize()

    # AND the error message should be exact and descriptive
    error_message = str(exc_info.value)
    assert error_message == (
        "Workflow validation error in InfiniteLoopWorkflow: " "Graph contains a self-edge (StartNode >> StartNode)."
    )


def test_workflow_serialization__node_with_conditional_loop_is_valid():
    """
    Tests that a node with conditional ports (one looping back, one going forward) is valid.
    """

    # GIVEN a workflow where a node has ports with one looping back to itself and one going to another node
    class State(BaseState):
        should_loop: bool = False

    class StartNode(BaseNode[State]):
        class Ports(BaseNode.Ports):
            loop = Port.on_if(State.should_loop.equals(True))
            end = Port.on_else()

    class EndNode(BaseNode[State]):
        pass

    class ConditionalLoopWorkflow(BaseWorkflow[BaseInputs, State]):
        graph = {
            StartNode.Ports.loop >> StartNode,
            StartNode.Ports.end >> EndNode,
        }

    # WHEN we attempt to serialize the workflow
    workflow_display = get_workflow_display(workflow_class=ConditionalLoopWorkflow)

    # THEN it should NOT raise a WorkflowValidationError
    result = workflow_display.serialize()
    assert result is not None



from deepdiff import DeepDiff

from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_tool_calling_node_inline_workflow.workflow import BasicToolCallingNodeInlineWorkflowWorkflow


def test_serialize_workflow():
    # GIVEN a Workflow that uses a generic node
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=BasicToolCallingNodeInlineWorkflowWorkflow)

    serialized_workflow: dict = workflow_display.serialize()
    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 1

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 2
    assert not DeepDiff(
        [
            {"id": "dbedc4ee-be3b-4135-8c26-3643c0b6a530", "key": "text", "type": "STRING"},
            {"id": "c5733df5-03bb-498e-a770-8ef9bff85df3", "key": "chat_history", "type": "CHAT_HISTORY"},
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    tool_calling_node = workflow_raw_data["nodes"][1]
    function_attributes = next(
        attribute for attribute in tool_calling_node["attributes"] if attribute["name"] == "functions"
    )
    assert function_attributes["id"] == "102d8447-5232-4e96-8192-9b1ca0f02650"
    assert function_attributes["name"] == "functions"
    assert function_attributes["value"]["type"] == "CONSTANT_VALUE"
    assert function_attributes["value"]["value"]["type"] == "JSON"
    inline_workflow_tool = function_attributes["value"]["value"]["value"][0]
    assert inline_workflow_tool["type"] == "INLINE_WORKFLOW"
    assert inline_workflow_tool["name"] == "BasicInlineSubworkflowWorkflow"
    assert (
        inline_workflow_tool["description"] == "\n    A workflow that gets the weather for a given city and date.\n    "
    )
    assert inline_workflow_tool["exec_config"]["input_variables"] == [
        {
            "id": "fa73da37-34c3-47a9-be58-69cc6cdbfca5",
            "key": "city",
            "type": "STRING",
            "default": None,
            "required": True,
            "extensions": {"color": None},
        },
        {
            "id": "aba1e6e0-dfa7-4c15-a4e6-aec6feebfaca",
            "key": "date",
            "type": "STRING",
            "default": None,
            "required": True,
            "extensions": {"color": None},
        },
    ]
    assert inline_workflow_tool["exec_config"]["output_variables"] == [
        {
            "id": "99afb757-2782-465d-ab55-80ccf50552b9",
            "key": "temperature",
            "type": "NUMBER",
        },
        {
            "id": "7444a019-081a-4e10-a528-3249299159f7",
            "key": "reasoning",
            "type": "STRING",
        },
    ]
    assert inline_workflow_tool["exec_config"]["state_variables"] == []
    inline_workflow_data = inline_workflow_tool["exec_config"]["workflow_raw_data"]
    assert inline_workflow_data["display_data"] == {"viewport": {"x": 0.0, "y": 0.0, "zoom": 1.0}}
    assert inline_workflow_data["definition"] == {
        "name": "BasicInlineSubworkflowWorkflow",
        "module": [
            "tests",
            "workflows",
            "basic_tool_calling_node_inline_workflow",
            "workflow",
        ],
    }
    assert inline_workflow_data["output_values"] == [
        {
            "output_variable_id": "99afb757-2782-465d-ab55-80ccf50552b9",
            "value": {
                "type": "NODE_OUTPUT",
                "node_id": "07a83d1a-7948-4a23-9f46-9a60382d3a48",
                "node_output_id": "272eccdb-6903-4f32-8159-8bfb87c65c2e",
            },
        },
        {
            "output_variable_id": "7444a019-081a-4e10-a528-3249299159f7",
            "value": {
                "type": "NODE_OUTPUT",
                "node_id": "07a83d1a-7948-4a23-9f46-9a60382d3a48",
                "node_output_id": "8616509d-fecb-4bbf-afdc-bd4c5de35ce7",
            },
        },
    ]
    inner_start_node = next(
        node for node in inline_workflow_data["nodes"] if node["type"] == "GENERIC" and node["label"] == "Start Node"
    )
    assert inner_start_node == {
        "id": "07a83d1a-7948-4a23-9f46-9a60382d3a48",
        "label": "Start Node",
        "type": "GENERIC",
        "should_file_merge": True,
        "display_data": {"position": {"x": 200.0, "y": -50.0}},
        "base": {
            "name": "BaseNode",
            "module": ["vellum", "workflows", "nodes", "bases", "base"],
        },
        "definition": {
            "name": "StartNode",
            "module": [
                "tests",
                "workflows",
                "basic_tool_calling_node_inline_workflow",
                "workflow",
            ],
        },
        "trigger": {
            "id": "8a7854a7-a595-44ce-8197-1f9cd2c78f10",
            "merge_behavior": "AWAIT_ATTRIBUTES",
        },
        "ports": [
            {
                "id": "ee444f1a-b126-4480-8eb2-d22506640763",
                "name": "default",
                "type": "DEFAULT",
            }
        ],
        "adornments": None,
        "attributes": [
            {
                "id": "b5e2d415-4368-43c1-826c-ff6511bf6942",
                "name": "city",
                "value": {
                    "type": "WORKFLOW_INPUT",
                    "input_variable_id": "fa73da37-34c3-47a9-be58-69cc6cdbfca5",
                },
            },
            {
                "id": "5b3ed6d0-8cd6-41b6-ad55-d380b41f943b",
                "name": "date",
                "value": {
                    "type": "WORKFLOW_INPUT",
                    "input_variable_id": "aba1e6e0-dfa7-4c15-a4e6-aec6feebfaca",
                },
            },
        ],
        "outputs": [
            {
                "id": "272eccdb-6903-4f32-8159-8bfb87c65c2e",
                "name": "temperature",
                "type": "NUMBER",
                "value": None,
                "schema": {"type": "number"},
            },
            {
                "id": "8616509d-fecb-4bbf-afdc-bd4c5de35ce7",
                "name": "reasoning",
                "type": "STRING",
                "value": None,
                "schema": {"type": "string"},
            },
        ],
    }
